import json
import pathlib
from abc import ABC
from typing import Any, Dict, Literal, Type

import pyperf


def load_data():
    with open(pathlib.Path(__file__).parent / "data" / "issue.json") as f:
        return json.load(f)


class AbstractBenchmark(ABC):
    LIBRARY: str

    def __init__(self, runner: pyperf.Runner) -> None:
        self.runner = runner

    def warmup(self, data: Dict[str, Any]) -> None:
        pass

    def run_loader(self, data: Dict[str, Any]) -> pyperf.Benchmark:
        pass

    def run_dumper(self, data: Dict[str, Any]) -> pyperf.Benchmark:
        pass

    def _bench_loader_func(self, func, *args, **kwargs) -> pyperf.Benchmark:
        return self.runner.bench_func(
            self.get_name("load"), func, *args, **kwargs
        )

    def _bench_dumper_func(self, func, *args, **kwargs) -> pyperf.Benchmark:
        return self.runner.bench_func(
            self.get_name("dump"), func, *args, **kwargs
        )

    def get_name(self, bench_type: Literal["dump", "load"]) -> str:
        return f"{self.LIBRARY}[{bench_type}]"


class BenchmarkRunner:
    def __init__(self, benchmark_cls: Type[AbstractBenchmark]) -> None:
        self._runner = pyperf.Runner()
        self._benchmark = benchmark_cls(self._runner)
        self._data = load_data()

    def run(self, benchmark_type: str) -> None:
        self._benchmark.warmup(self._data)
        if benchmark_type == "load":
            self._benchmark.run_loader(self._data)
        elif benchmark_type == "dump":
            self._benchmark.run_dumper(self._data)
        else:
            raise ValueError(f"Unknown benchmark_type: {benchmark_type}")



from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, Optional, no_type_check

import pytest

from mashumaro import DataClassDictMixin, field_options, pass_through
from mashumaro.config import ADD_SERIALIZATION_CONTEXT, BaseConfig
from mashumaro.exceptions import BadHookSignature


class BaseClassWithSerializationContext(DataClassDictMixin):
    class Config(BaseConfig):
        code_generation_options = [ADD_SERIALIZATION_CONTEXT]


@dataclass
class Foo(BaseClassWithSerializationContext):
    baz: int

    class Config(BaseConfig):
        code_generation_options = []

    def __pre_serialize__(self):
        return self

    def __post_serialize__(self, d: Dict):
        return d


@dataclass
class Bar(BaseClassWithSerializationContext):
    baz: int

    def __pre_serialize__(self, context: Optional[Dict] = None):
        return self

    def __post_serialize__(self, d: Dict, context: Optional[Dict] = None):
        if context and context.get("omit_baz"):
            d.pop("baz")
        return d


@dataclass
class FooBarBaz(BaseClassWithSerializationContext):
    foo: Foo
    bar: Bar
    baz: int

    def __pre_serialize__(self, context: Optional[Dict] = None):
        return self

    def __post_serialize__(self, d: Dict, context: Optional[Dict] = None):
        if context and context.get("omit_baz"):
            d.pop("baz")
        return d


def test_bad_pre_deserialize_hook():
    with pytest.raises(BadHookSignature):

        class DataClass(DataClassDictMixin):
            x: int

            @no_type_check
            def __pre_deserialize__(
                self, d: Dict[Any, Any]
            ) -> Dict[Any, Any]: ...


def test_bad_post_deserialize_hook():
    with pytest.raises(BadHookSignature):

        class DataClass(DataClassDictMixin):
            x: int

            @no_type_check
            def __post_deserialize__(
                self, obj: "DataClass"
            ) -> "DataClass": ...


def test_pre_deserialize_hook():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int

        @classmethod
        def __pre_deserialize__(cls, d: Dict[Any, Any]) -> Dict[Any, Any]:
            return {k.lower(): v for k, v in d.items()}

    assert DataClass.from_dict({"X": 123}) == DataClass(x=123)


def test_post_deserialize_hook():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int

        @classmethod
        def __post_deserialize__(cls, obj: "DataClass") -> "DataClass":
            obj.x = 456
            return obj

    assert DataClass.from_dict({"x": 123}) == DataClass(x=456)


def test_pre_serialize_hook():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int
        counter: ClassVar[int] = 0

        def __pre_serialize__(self) -> "DataClass":
            DataClass.counter += 1
            return self

    instance = DataClass(x=123)
    assert instance.to_dict() == {"x": 123}
    assert instance.counter == 1


def test_post_serialize_hook():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int
        counter: ClassVar[int] = 0

        def __post_serialize__(self, d: Dict[Any, Any]) -> Dict[Any, Any]:
            return {k.upper(): v for k, v in d.items()}

    instance = DataClass(x=123)
    assert instance.to_dict() == {"X": 123}


def test_hook_stub_is_not_called(mocker):
    pre_deserialize_hook = mocker.patch.object(
        DataClassDictMixin, "__pre_deserialize__"
    )
    post_deserialize_hook = mocker.patch.object(
        DataClassDictMixin, "__post_deserialize__"
    )
    pre_serialize_hook = mocker.patch.object(
        DataClassDictMixin, "__pre_serialize__"
    )
    post_serialize_hook = mocker.patch.object(
        DataClassDictMixin, "__post_serialize__"
    )

    @dataclass
    class A(DataClassDictMixin):
        a: int

    obj = A.from_dict({"a": 1})
    obj.to_dict()

    pre_deserialize_hook.assert_not_called()
    post_deserialize_hook.assert_not_called()
    pre_serialize_hook.assert_not_called()
    post_serialize_hook.assert_not_called()


def test_hook_in_parent_class(mocker):
    class A:
        @classmethod
        def __pre_deserialize__(cls, d):
            return d  # pragma: no cover

        @classmethod
        def __post_deserialize__(cls, obj):
            return obj  # pragma: no cover

        def __pre_serialize__(self):
            return self  # pragma: no cover

        def __post_serialize__(self, d):
            return d  # pragma: no cover

    @dataclass
    class B(A, DataClassDictMixin):
        a: int

    pre_deserialize_hook = mocker.patch.object(A, "__pre_deserialize__")
    post_deserialize_hook = mocker.patch.object(A, "__post_deserialize__")
    pre_serialize_hook = mocker.patch.object(
        A, "__pre_serialize__", return_value=B(a=1)
    )
    post_serialize_hook = mocker.patch.object(A, "__post_serialize__")

    B.from_dict({"a": 1})
    B(a=1).to_dict()

    pre_deserialize_hook.assert_called_once()
    post_deserialize_hook.assert_called_once()
    pre_serialize_hook.assert_called_once()
    post_serialize_hook.assert_called_once()


def test_post_deserialize_hook_with_pass_through_field():
    @dataclass
    class MyClass(DataClassDictMixin):
        x: int = field(metadata=field_options(deserialize=pass_through))

        @classmethod
        def __post_deserialize__(cls, obj):
            return obj

    assert MyClass.from_dict({"x": 42}) == MyClass(42)


def test_post_deserialize_hook_with_empty_dataclass():
    @dataclass
    class MyClass(DataClassDictMixin):
        @classmethod
        def __post_deserialize__(cls, obj):
            return obj  # pragma: no cover

    assert MyClass.from_dict({}) == MyClass()


def test_passing_context_into_hook():
    foo = FooBarBaz(foo=Foo(1), bar=Bar(baz=2), baz=3)
    assert foo.to_dict() == {"foo": {"baz": 1}, "bar": {"baz": 2}, "baz": 3}
    assert foo.to_dict(context={"omit_baz": True}) == {
        "foo": {"baz": 1},
        "bar": {},
    }



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar

from mashumaro.mixins.json import DataClassJSONMixin

T = TypeVar("T")


@dataclass
class Foo(Generic[T], DataClassJSONMixin):
    x: T


@dataclass
class Bar(Foo[int]):
    pass



from dataclasses import dataclass
from datetime import date
from itertools import permutations
from typing import Any, Dict, List, Union

import pytest

from mashumaro import DataClassDictMixin
from mashumaro.codecs.basic import encode
from tests.utils import same_types


@dataclass
class UnionTestCase:
    type: Any
    dumped: Any
    loaded: Any


@pytest.mark.parametrize(
    "test_case",
    [
        # int | str
        UnionTestCase(Union[int, str], 1, 1),
        UnionTestCase(Union[int, str], "1", "1"),
        UnionTestCase(Union[int, str], 1.0, 1),
        UnionTestCase(Union[int, str], 1.2, 1),
        UnionTestCase(Union[int, str], "a", "a"),
        UnionTestCase(Union[int, str], [1, 2], "[1, 2]"),
        # str | int
        UnionTestCase(Union[str, int], 1, 1),
        UnionTestCase(Union[str, int], "1", "1"),
        UnionTestCase(Union[str, int], 1.0, "1.0"),
        UnionTestCase(Union[str, int], 1.2, "1.2"),
        UnionTestCase(Union[str, int], "a", "a"),
        UnionTestCase(Union[str, int], [1, 2], "[1, 2]"),
        # str | date
        UnionTestCase(Union[str, date], "2024-11-12", "2024-11-12"),
        UnionTestCase(Union[str, date], "a", "a"),
        # date | str
        UnionTestCase(Union[date, str], "2024-11-12", date(2024, 11, 12)),
        UnionTestCase(Union[date, str], "a", "a"),
        # int | float
        UnionTestCase(Union[int, float], 1, 1),
        UnionTestCase(Union[int, float], 1.0, 1.0),
        UnionTestCase(Union[int, float], "1", 1),
        UnionTestCase(Union[int, float], "1.0", 1.0),
        # float | int
        UnionTestCase(Union[float, int], 1, 1),
        UnionTestCase(Union[float, int], 1.0, 1.0),
        UnionTestCase(Union[float, int], "1", 1.0),
        UnionTestCase(Union[float, int], "1.0", 1.0),
        # bool | int
        UnionTestCase(Union[bool, int], 1, 1),
        UnionTestCase(Union[bool, int], 1.0, True),
        UnionTestCase(Union[bool, int], True, True),
        UnionTestCase(Union[bool, int], "1", True),
        UnionTestCase(Union[bool, int], "1.2", True),
        UnionTestCase(Union[bool, int], [1, 2], True),
        UnionTestCase(Union[bool, int], "a", True),
        # int | bool
        UnionTestCase(Union[int, bool], 1, 1),
        UnionTestCase(Union[int, bool], 1.0, 1),
        UnionTestCase(Union[int, bool], True, True),
        UnionTestCase(Union[int, bool], "1", 1),
        UnionTestCase(Union[int, bool], "1.2", True),
        UnionTestCase(Union[int, bool], [1, 2], True),
        UnionTestCase(Union[int, bool], "a", True),
        # dict[int, int] | list[int]
        UnionTestCase(Union[Dict[int, int], List[int]], {1: 2}, {1: 2}),
        UnionTestCase(Union[Dict[int, int], List[int]], {"1": "2"}, {1: 2}),
        UnionTestCase(Union[Dict[int, int], List[int]], [1], [1]),
        UnionTestCase(Union[Dict[int, int], List[int]], ["1"], [1]),
        # str | list[str]
        UnionTestCase(Union[str, List[str]], ["a"], ["a"]),
        UnionTestCase(Union[str, List[str]], "abc", "abc"),
        UnionTestCase(Union[str, List[str]], 1, "1"),
        UnionTestCase(Union[str, List[str]], [1, 2], ["1", "2"]),
        # int | float | None
        UnionTestCase(Union[int, float, None], None, None),
        UnionTestCase(Union[int, float, None], 1, 1),
        UnionTestCase(Union[int, float, None], 1.2, 1.2),
        UnionTestCase(Union[int, float, None], "1", 1),
        UnionTestCase(Union[int, float, None], "1.2", 1.2),
        UnionTestCase(Union[int, float, None], "a", None),
        # None | int | float
        UnionTestCase(Union[None, int, float], None, None),
        UnionTestCase(Union[None, int, float], 1, 1),
        UnionTestCase(Union[None, int, float], 1.2, 1.2),
        UnionTestCase(Union[None, int, float], "1", None),
        UnionTestCase(Union[None, int, float], "1.2", None),
        UnionTestCase(Union[None, int, float], "a", None),
        # int | None | float
        UnionTestCase(Union[int, None, float], None, None),
        UnionTestCase(Union[int, None, float], 1, 1),
        UnionTestCase(Union[int, None, float], 1.2, 1.2),
        UnionTestCase(Union[int, None, float], "1", 1),
        UnionTestCase(Union[int, None, float], "1.2", None),
        UnionTestCase(Union[int, None, float], "a", None),
    ],
)
def test_union_deserialization(test_case):
    @dataclass
    class DataClass(DataClassDictMixin):
        x: test_case.type

    instance = DataClass(x=test_case.loaded)
    loaded = DataClass.from_dict({"x": test_case.dumped})
    assert loaded == instance
    assert same_types(loaded.x, instance.x)


@pytest.mark.parametrize(
    "test_case",
    [
        UnionTestCase(Union[int, str], 1, 1),
        UnionTestCase(Union[int, str], "1", "1"),
        UnionTestCase(Union[int, str], "a", "a"),
        UnionTestCase(Union[Dict[int, int], List[int]], {1: 2}, {1: 2}),
        UnionTestCase(Union[Dict[int, int], List[int]], [1], [1]),
        UnionTestCase(Union[str, List[str]], ["a"], ["a"]),
        UnionTestCase(Union[str, List[str]], "abc", "abc"),
    ],
)
def test_union_serialization(test_case):
    @dataclass
    class DataClass(DataClassDictMixin):
        x: test_case.type

    instance = DataClass(x=test_case.loaded)
    dumped = instance.to_dict()
    assert dumped == {"x": test_case.dumped}
    assert same_types(dumped["x"], test_case.dumped)


def test_union_encoding():
    for variants in permutations((int, float, str, bool)):
        for value in (1, 2.0, 3.1, "4", "5.0", True, False):
            encoded = encode(value, Union[variants])
            assert value == encoded
            assert same_types(value, encoded)



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional, Union

import cattr
import cattrs.gen
import pyperf

from benchmark.common import AbstractBenchmark


class IssueState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class MilestoneState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class IssueStateReason(Enum):
    COMPLETED = "completed"
    REOPENED = "reopened"
    NOT_PLANNED = "not_planned"


class AuthorAssociation(Enum):
    COLLABORATOR = "COLLABORATOR"
    CONTRIBUTOR = "CONTRIBUTOR"
    FIRST_TIMER = "FIRST_TIMER"
    FIRST_TIME_CONTRIBUTOR = "FIRST_TIME_CONTRIBUTOR"
    MANNEQUIN = "MANNEQUIN"
    MEMBER = "MEMBER"
    NONE = "NONE"
    OWNER = "OWNER"


@dataclass(slots=True)
class User:
    id: int
    node_id: str
    avatar_url: str
    gravatar_id: Optional[str]
    url: str
    html_url: str
    followers_url: str
    following_url: str
    gists_url: str
    starred_url: str
    subscriptions_url: str
    organizations_url: str
    repos_url: str
    events_url: str
    received_events_url: str
    type: str
    site_admin: bool
    name: Optional[str] = None
    email: Optional[str] = None
    starred_at: Optional[datetime] = None


@dataclass(slots=True)
class IssueLabel:
    id: int
    node_id: str
    url: str
    name: str
    description: Optional[str]
    color: Optional[str]
    default: bool


@dataclass(slots=True)
class Milestone:
    url: str
    html_url: str
    labels_url: str
    id: int
    node_id: str
    number: int
    title: str
    description: Optional[str]
    creator: Optional[User]
    open_issues: int
    closed_issues: int
    created_at: datetime
    updated_at: datetime
    closed_at: Optional[datetime]
    due_on: Optional[datetime]
    state: MilestoneState = MilestoneState.OPEN


@dataclass(slots=True)
class Reactions:
    url: str
    total_count: int
    plus_one: int
    minus_one: int
    laugh: int
    confused: int
    heart: int
    hooray: int
    eyes: int
    rocket: int


@dataclass(slots=True)
class Issue:
    id: int
    node_id: str
    url: str
    repository_url: str
    labels_url: str
    comments_url: str
    events_url: str
    html_url: str
    number: int
    state: IssueState
    state_reason: Optional[IssueStateReason]
    title: str
    body: Optional[str]
    user: Optional[User]
    labels: list[Union[IssueLabel, str]]
    assignee: Optional[User]
    assignees: Optional[list[User]]
    milestone: Optional[Milestone]
    locked: bool
    active_lock_reason: Optional[str]
    comments: int
    closed_at: Optional[datetime]
    created_at: datetime
    updated_at: datetime
    closed_by: Optional[User]
    author_association: AuthorAssociation
    draft: bool = False
    body_html: Optional[str] = None
    body_text: Optional[str] = None
    timeline_url: Optional[str] = None
    reactions: Optional[Reactions] = None


class Benchmark(AbstractBenchmark):
    LIBRARY = "cattrs"

    def __init__(self, runner: pyperf.Runner):
        super().__init__(runner)
        self._converter = cattr.Converter(detailed_validation=False)
        self._converter.register_structure_hook(
            Reactions,
            cattr.gen.make_dict_structure_fn(
                Reactions,
                self._converter,
                plus_one=cattrs.gen.override(rename="+1"),
                minus_one=cattrs.gen.override(rename="-1"),
            ),
        )
        self._converter.register_unstructure_hook(
            Reactions,
            cattr.gen.make_dict_unstructure_fn(
                Reactions,
                self._converter,
                plus_one=cattrs.gen.override(rename="+1"),
                minus_one=cattrs.gen.override(rename="-1"),
            ),
        )
        self._converter.register_structure_hook(
            datetime, lambda o, _: datetime.fromisoformat(o)
        )
        self._converter.register_unstructure_hook(
            datetime, lambda o: o.isoformat()
        )
        self._converter.register_structure_hook(
            Union[IssueLabel, str],
            lambda o, _: self._converter.structure(
                o, IssueLabel if isinstance(o, dict) else str
            ),
        )

    def warmup(self, data: Dict[str, Any]):
        self._converter.unstructure(self._converter.structure(data, Issue))

    def run_loader(self, data: Dict[str, Any]) -> pyperf.Benchmark:
        return self._bench_loader_func(self._converter.structure, data, Issue)

    def run_dumper(self, data: Dict[str, Any]) -> pyperf.Benchmark:
        obj = self._converter.structure(data, Issue)
        return self._bench_dumper_func(self._converter.unstructure, obj)



from typing import ChainMap, Deque, List, Mapping, Set, Tuple


def same_types(first, second):
    if isinstance(first, (List, Deque, Tuple)):
        return all(map(lambda x: same_types(*x), zip(first, second)))
    elif isinstance(first, ChainMap):
        return all(map(lambda x: same_types(*x), zip(first.maps, second.maps)))
    elif isinstance(first, Mapping):
        return all(
            map(lambda x: same_types(*x), zip(first.keys(), second.keys()))
        ) and all(
            map(lambda x: same_types(*x), zip(first.values(), second.values()))
        )
    elif isinstance(first, Set):
        return all(
            map(lambda x: same_types(*x), zip(sorted(first), sorted(second)))
        )
    else:
        return type(first) is type(second)



import collections
import enum
import typing
from dataclasses import dataclass, field
from datetime import date, datetime
from typing import (
    FrozenSet,
    Generic,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import pytest
from typing_extensions import TypedDict

from mashumaro import DataClassDictMixin, pass_through
from mashumaro.config import ADD_DIALECT_SUPPORT, BaseConfig
from mashumaro.dialect import Dialect
from mashumaro.exceptions import BadDialect
from mashumaro.mixins.msgpack import DataClassMessagePackMixin
from mashumaro.mixins.msgpack import default_encoder as msgpack_encoder
from mashumaro.types import SerializationStrategy

from .conftest import add_unpack_method

T_Date = TypeVar("T_Date", bound=date)


class HexSerializationStrategy(SerializationStrategy):
    def serialize(self, value: int) -> str:
        return hex(value)

    def deserialize(self, value: str) -> int:
        return int(value, 16)


class OrdinalDialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": date.toordinal,
            "deserialize": date.fromordinal,
        },
        int: HexSerializationStrategy(),
    }


class FormattedDialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": lambda dt: dt.strftime("%Y/%m/%d"),
            "deserialize": lambda s: datetime.strptime(s, "%Y/%m/%d").date(),
        },
        int: HexSerializationStrategy(),
    }


class ISODialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": date.isoformat,
            "deserialize": date.fromisoformat,
        },
        int: HexSerializationStrategy(),
    }


class EmptyDialect(Dialect):
    pass


class OmitNoneDialect(Dialect):
    omit_none = True


class NotOmitNoneDialect(Dialect):
    omit_none = False


class OmitDefaultDialect(Dialect):
    omit_default = True


class NotOmitDefaultDialect(Dialect):
    omit_default = False


@dataclass
class DataClassWithoutDialects(DataClassDictMixin):
    dt: date
    i: int


@dataclass
class DataClassWithDefaultDialect(DataClassDictMixin):
    dt: date
    i: int

    class Config(BaseConfig):
        dialect = OrdinalDialect


@dataclass
class DataClassWithDialectSupport(DataClassDictMixin):
    dt: date
    i: int

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class BaseEntityWithDialect(DataClassDictMixin):
    class Config:
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class Entity1(BaseEntityWithDialect):
    dt1: date


@dataclass
class Entity2(BaseEntityWithDialect):
    dt2: date


@dataclass
class DataClassWithDialectSupportAndDefaultDialect(DataClassDictMixin):
    dt: date
    i: int

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = FormattedDialect


@dataclass
class GenericDataClassWithoutDialects(Generic[T_Date], DataClassDictMixin):
    dt: T_Date
    i: int


@dataclass
class GenericDataClassWithDefaultDialect(Generic[T_Date], DataClassDictMixin):
    dt: T_Date
    i: int

    class Config(BaseConfig):
        dialect = OrdinalDialect


@dataclass
class GenericDataClassWithDialectSupport(Generic[T_Date], DataClassDictMixin):
    dt: T_Date
    i: int

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class GenericDataClassWithDialectSupportAndDefaultDialect(
    Generic[T_Date], DataClassDictMixin
):
    dt: T_Date
    i: int

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]
        dialect = FormattedDialect


class MyNamedTuple(NamedTuple):
    x: DataClassWithDialectSupport = DataClassWithDialectSupport(
        dt=date(2022, 1, 1),
        i=999,
    )
    y: DataClassWithoutDialects = DataClassWithoutDialects(
        dt=date(2022, 1, 1),
        i=999,
    )


class MyTypedDict(TypedDict):
    x: DataClassWithDialectSupport
    y: DataClassWithoutDialects


@dataclass
class DataClassWithNamedTupleWithDialectSupport(DataClassDictMixin):
    x: MyNamedTuple

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class DataClassWithNamedTupleWithoutDialectSupport(DataClassDictMixin):
    x: MyNamedTuple


@dataclass
class DataClassWithTypedDictWithDialectSupport(DataClassDictMixin):
    x: MyTypedDict

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class DataClassWithTypedDictWithoutDialectSupport(DataClassDictMixin):
    x: MyTypedDict


@dataclass
class DataClassWithUnionWithDialectSupport(DataClassDictMixin):
    x: List[Union[DataClassWithDialectSupport, DataClassWithoutDialects]]

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class MessagePackDataClass(DataClassMessagePackMixin):
    b_1: bytes
    b_2: bytearray
    dep_1: DataClassWithoutDialects
    dep_2: GenericDataClassWithoutDialects[date]


@dataclass
class DataClassWithOptionalAndOmitNoneDialect(DataClassDictMixin):
    x: Optional[int] = None

    class Config(BaseConfig):
        dialect = OmitNoneDialect


@dataclass
class DataClassWithOptionalAndOmitNoneDialectAndOmitNoneFalse(
    DataClassDictMixin
):
    x: Optional[int] = None

    class Config(BaseConfig):
        dialect = OmitNoneDialect
        omit_none = False


@dataclass
class DataClassWithOptionalAndNotOmitNoneDialectAndOmitNoneTrue(
    DataClassDictMixin
):
    x: Optional[int] = None

    class Config(BaseConfig):
        dialect = NotOmitNoneDialect
        omit_none = True


@dataclass
class DataClassWithOptionalAndEmptyDialect(DataClassDictMixin):
    x: Optional[int] = None

    class Config(BaseConfig):
        dialect = EmptyDialect
        omit_none = True


@dataclass
class DataClassWithOptionalAndDialectSupport(DataClassDictMixin):
    x: Optional[int] = None

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


class MyStrEnum(str, enum.Enum):
    VAL1 = "val1"
    VAL2 = "val2"


@dataclass
class DataClassWithDefaultAndOmitDefaultDialect(DataClassDictMixin):
    a: int = 42
    b: float = 3.14
    c: str = "foo"
    d: bool = False
    e: None = None
    f: type(None) = None
    g: Tuple[int] = (1, 2, 3)
    h: MyNamedTuple = field(default_factory=lambda: MyNamedTuple())
    i: List[int] = field(default_factory=list)
    j: List[int] = field(default_factory=lambda: [1, 2, 3])
    k: Set[int] = field(default_factory=set)
    l: Set[int] = field(default_factory=lambda: {1, 2, 3})
    m: FrozenSet[int] = field(default_factory=frozenset)
    n: FrozenSet[int] = field(default_factory=lambda: frozenset({1, 2, 3}))
    o: str = MyStrEnum.VAL1

    class Config(BaseConfig):
        dialect = OmitDefaultDialect


@dataclass
class DataClassWithDefaultAndOmitDefaultDialectAndOmitDefaultFalse(
    DataClassDictMixin
):
    a: int = 42
    b: float = 3.14
    c: str = "foo"
    d: bool = False
    e: None = None
    f: type(None) = None
    g: Tuple[int] = (1, 2, 3)
    h: MyNamedTuple = field(default_factory=lambda: MyNamedTuple())
    i: List[int] = field(default_factory=list)
    j: List[int] = field(default_factory=lambda: [1, 2, 3])
    k: Set[int] = field(default_factory=set)
    l: Set[int] = field(default_factory=lambda: {1, 2, 3})
    m: FrozenSet[int] = field(default_factory=frozenset)
    n: FrozenSet[int] = field(default_factory=lambda: frozenset({1, 2, 3}))
    o: str = MyStrEnum.VAL1

    class Config(BaseConfig):
        dialect = OmitDefaultDialect
        omit_default = False


@dataclass
class DataClassWithDefaultAndNotOmitDefaultDialectAndOmitDefaultTrue(
    DataClassDictMixin
):
    a: int = 42
    b: float = 3.14
    c: str = "foo"
    d: bool = False
    e: None = None
    f: type(None) = None
    g: Tuple[int] = (1, 2, 3)
    h: MyNamedTuple = field(default_factory=lambda: MyNamedTuple())
    i: List[int] = field(default_factory=list)
    j: List[int] = field(default_factory=lambda: [1, 2, 3])
    k: Set[int] = field(default_factory=set)
    l: Set[int] = field(default_factory=lambda: {1, 2, 3})
    m: FrozenSet[int] = field(default_factory=frozenset)
    n: FrozenSet[int] = field(default_factory=lambda: frozenset({1, 2, 3}))
    o: str = MyStrEnum.VAL1

    class Config(BaseConfig):
        dialect = NotOmitDefaultDialect
        omit_default = True


@dataclass
class DataClassWithDefaultAndEmptyDialect(DataClassDictMixin):
    a: int = 42
    b: float = 3.14
    c: str = "foo"
    d: bool = False
    e: None = None
    f: type(None) = None
    g: Tuple[int] = (1, 2, 3)
    h: MyNamedTuple = field(default_factory=lambda: MyNamedTuple())
    i: List[int] = field(default_factory=list)
    j: List[int] = field(default_factory=lambda: [1, 2, 3])
    k: Set[int] = field(default_factory=set)
    l: Set[int] = field(default_factory=lambda: {1, 2, 3})
    m: FrozenSet[int] = field(default_factory=frozenset)
    n: FrozenSet[int] = field(default_factory=lambda: frozenset({1, 2, 3}))
    o: str = MyStrEnum.VAL1

    class Config(BaseConfig):
        dialect = EmptyDialect
        omit_default = True


@dataclass
class DataClassWithDefaultAndDialectSupport(DataClassDictMixin):
    a: int = 42
    b: float = 3.14
    c: str = "foo"
    d: bool = False
    e: None = None
    f: type(None) = None
    g: Tuple[int] = (1, 2, 3)
    h: MyNamedTuple = field(default_factory=lambda: MyNamedTuple())
    i: List[int] = field(default_factory=list)
    j: List[int] = field(default_factory=lambda: [1, 2, 3])
    k: Set[int] = field(default_factory=set)
    l: Set[int] = field(default_factory=lambda: {1, 2, 3})
    m: FrozenSet[int] = field(default_factory=frozenset)
    n: FrozenSet[int] = field(default_factory=lambda: frozenset({1, 2, 3}))
    o: str = MyStrEnum.VAL1

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


def test_default_dialect():
    dt = date.today()
    ordinal = dt.toordinal()
    obj = DataClassWithDefaultDialect(dt, 255)
    assert obj.to_dict() == {"dt": ordinal, "i": "0xff"}
    assert (
        DataClassWithDefaultDialect.from_dict({"dt": ordinal, "i": "0xff"})
        == obj
    )
    with pytest.raises(TypeError):
        obj.to_dict(dialect=OrdinalDialect)


def test_dialect():
    dt = date.today()
    ordinal = dt.toordinal()
    obj = DataClassWithDialectSupport(dt, 255)
    assert obj.to_dict(dialect=OrdinalDialect) == {"dt": ordinal, "i": "0xff"}
    assert (
        DataClassWithDialectSupport.from_dict(
            {"dt": ordinal, "i": "0xff"}, dialect=OrdinalDialect
        )
        == obj
    )


def test_dialect_with_default():
    dt = date.today()
    ordinal = dt.toordinal()
    formatted = dt.strftime("%Y/%m/%d")
    obj = DataClassWithDialectSupportAndDefaultDialect(dt, 255)
    assert obj.to_dict() == {"dt": formatted, "i": "0xff"}
    assert (
        DataClassWithDialectSupportAndDefaultDialect.from_dict(
            {"dt": formatted, "i": "0xff"}
        )
        == obj
    )
    assert obj.to_dict(dialect=None) == {"dt": formatted, "i": "0xff"}
    assert (
        DataClassWithDialectSupportAndDefaultDialect.from_dict(
            {"dt": formatted, "i": "0xff"}, dialect=None
        )
        == obj
    )
    assert obj.to_dict(dialect=OrdinalDialect) == {"dt": ordinal, "i": "0xff"}
    assert (
        DataClassWithDialectSupportAndDefaultDialect.from_dict(
            {"dt": ordinal, "i": "0xff"}, dialect=OrdinalDialect
        )
        == obj
    )


def test_bad_default_dialect():
    with pytest.raises(BadDialect):

        @dataclass
        class _(DataClassDictMixin):
            dt: date

            class Config(BaseConfig):
                dialect = int

    with add_unpack_method:
        with pytest.raises(BadDialect):

            @dataclass
            class _(DataClassDictMixin):
                dt: date

                class Config(BaseConfig):
                    dialect = int


def test_bad_dialect():
    dt = date.today()
    ordinal = dt.toordinal()
    obj = DataClassWithDialectSupport(dt, 255)
    with pytest.raises(BadDialect):
        DataClassWithDialectSupport.from_dict(
            {"dt": ordinal, "i": "0xff"}, dialect=int
        )
    with pytest.raises(BadDialect):
        obj.to_dict(dialect=int)


def test_inner_without_dialects():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: DataClassWithoutDialects
        inners: List[DataClassWithoutDialects]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    formatted = dt.strftime("%Y/%m/%d")
    ordinal = dt.toordinal()
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=DataClassWithoutDialects(dt, 255),
        inners=[DataClassWithoutDialects(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": iso, "i": 255},
        "inners": [{"dt": iso, "i": 255}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=OrdinalDialect) == {
        "dt": ordinal,
        "inner": {"dt": iso, "i": 255},
        "inners": [{"dt": iso, "i": 255}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": iso, "i": 255},
                "inners": [{"dt": iso, "i": 255}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": ordinal,
                "inner": {"dt": iso, "i": 255},
                "inners": [{"dt": iso, "i": 255}],
                "i": "0xff",
            },
            dialect=OrdinalDialect,
        )
        == obj
    )


def test_inner_with_default_dialect():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: DataClassWithDefaultDialect
        inners: List[DataClassWithDefaultDialect]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    formatted = dt.strftime("%Y/%m/%d")
    ordinal = dt.toordinal()
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=DataClassWithDefaultDialect(dt, 255),
        inners=[DataClassWithDefaultDialect(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": ordinal, "i": "0xff"},
        "inners": [{"dt": ordinal, "i": "0xff"}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=ISODialect) == {
        "dt": iso,
        "inner": {"dt": ordinal, "i": "0xff"},
        "inners": [{"dt": ordinal, "i": "0xff"}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": ordinal, "i": "0xff"},
                "inners": [{"dt": ordinal, "i": "0xff"}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": iso,
                "inner": {"dt": ordinal, "i": "0xff"},
                "inners": [{"dt": ordinal, "i": "0xff"}],
                "i": "0xff",
            },
            dialect=ISODialect,
        )
        == obj
    )


def test_inner_with_dialect_support():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: DataClassWithDialectSupport
        inners: List[DataClassWithDialectSupport]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    ordinal = dt.toordinal()
    formatted = dt.strftime("%Y/%m/%d")
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=DataClassWithDialectSupport(dt, 255),
        inners=[DataClassWithDialectSupport(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": iso, "i": 255},
        "inners": [{"dt": iso, "i": 255}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=OrdinalDialect) == {
        "dt": ordinal,
        "inner": {"dt": ordinal, "i": "0xff"},
        "inners": [{"dt": ordinal, "i": "0xff"}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": iso, "i": 255},
                "inners": [{"dt": iso, "i": 255}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": ordinal,
                "inner": {"dt": ordinal, "i": "0xff"},
                "inners": [{"dt": ordinal, "i": "0xff"}],
                "i": "0xff",
            },
            dialect=OrdinalDialect,
        )
        == obj
    )


def test_inner_with_dialect_support_and_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: DataClassWithDialectSupportAndDefaultDialect
        inners: List[DataClassWithDialectSupportAndDefaultDialect]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    formatted = dt.strftime("%Y/%m/%d")
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=DataClassWithDialectSupportAndDefaultDialect(dt, 255),
        inners=[DataClassWithDialectSupportAndDefaultDialect(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": formatted, "i": "0xff"},
        "inners": [{"dt": formatted, "i": "0xff"}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=ISODialect) == {
        "dt": iso,
        "inner": {"dt": iso, "i": "0xff"},
        "inners": [{"dt": iso, "i": "0xff"}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": formatted, "i": "0xff"},
                "inners": [{"dt": formatted, "i": "0xff"}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": iso,
                "inner": {"dt": iso, "i": "0xff"},
                "inners": [{"dt": iso, "i": "0xff"}],
                "i": "0xff",
            },
            dialect=ISODialect,
        )
        == obj
    )


def test_generic_without_dialects():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: GenericDataClassWithoutDialects[date]
        inners: List[GenericDataClassWithoutDialects[date]]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    formatted = dt.strftime("%Y/%m/%d")
    ordinal = dt.toordinal()
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=GenericDataClassWithoutDialects(dt, 255),
        inners=[GenericDataClassWithoutDialects(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": iso, "i": 255},
        "inners": [{"dt": iso, "i": 255}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=OrdinalDialect) == {
        "dt": ordinal,
        "inner": {"dt": iso, "i": 255},
        "inners": [{"dt": iso, "i": 255}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": iso, "i": 255},
                "inners": [{"dt": iso, "i": 255}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": ordinal,
                "inner": {"dt": iso, "i": 255},
                "inners": [{"dt": iso, "i": 255}],
                "i": "0xff",
            },
            dialect=OrdinalDialect,
        )
        == obj
    )


def test_generic_with_default_dialect():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: GenericDataClassWithDefaultDialect[date]
        inners: List[GenericDataClassWithDefaultDialect[date]]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    formatted = dt.strftime("%Y/%m/%d")
    ordinal = dt.toordinal()
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=GenericDataClassWithDefaultDialect(dt, 255),
        inners=[GenericDataClassWithDefaultDialect(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": ordinal, "i": "0xff"},
        "inners": [{"dt": ordinal, "i": "0xff"}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=ISODialect) == {
        "dt": iso,
        "inner": {"dt": ordinal, "i": "0xff"},
        "inners": [{"dt": ordinal, "i": "0xff"}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": ordinal, "i": "0xff"},
                "inners": [{"dt": ordinal, "i": "0xff"}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": iso,
                "inner": {"dt": ordinal, "i": "0xff"},
                "inners": [{"dt": ordinal, "i": "0xff"}],
                "i": "0xff",
            },
            dialect=ISODialect,
        )
        == obj
    )


def test_generic_with_dialect_support():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: GenericDataClassWithDialectSupport[date]
        inners: List[GenericDataClassWithDialectSupport[date]]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    ordinal = dt.toordinal()
    formatted = dt.strftime("%Y/%m/%d")
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=GenericDataClassWithDialectSupport(dt, 255),
        inners=[GenericDataClassWithDialectSupport(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": iso, "i": 255},
        "inners": [{"dt": iso, "i": 255}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=OrdinalDialect) == {
        "dt": ordinal,
        "inner": {"dt": ordinal, "i": "0xff"},
        "inners": [{"dt": ordinal, "i": "0xff"}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": iso, "i": 255},
                "inners": [{"dt": iso, "i": 255}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": ordinal,
                "inner": {"dt": ordinal, "i": "0xff"},
                "inners": [{"dt": ordinal, "i": "0xff"}],
                "i": "0xff",
            },
            dialect=OrdinalDialect,
        )
        == obj
    )


def test_generic_with_dialect_support_and_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date
        inner: GenericDataClassWithDialectSupportAndDefaultDialect[date]
        inners: List[GenericDataClassWithDialectSupportAndDefaultDialect[date]]
        i: int

        class Config(BaseConfig):
            code_generation_options = [ADD_DIALECT_SUPPORT]
            dialect = FormattedDialect

    dt = date.today()
    formatted = dt.strftime("%Y/%m/%d")
    iso = dt.isoformat()
    obj = DataClass(
        dt=dt,
        inner=GenericDataClassWithDialectSupportAndDefaultDialect(dt, 255),
        inners=[GenericDataClassWithDialectSupportAndDefaultDialect(dt, 255)],
        i=255,
    )
    assert obj.to_dict() == {
        "dt": formatted,
        "inner": {"dt": formatted, "i": "0xff"},
        "inners": [{"dt": formatted, "i": "0xff"}],
        "i": "0xff",
    }
    assert obj.to_dict(dialect=ISODialect) == {
        "dt": iso,
        "inner": {"dt": iso, "i": "0xff"},
        "inners": [{"dt": iso, "i": "0xff"}],
        "i": "0xff",
    }
    assert (
        DataClass.from_dict(
            {
                "dt": formatted,
                "inner": {"dt": formatted, "i": "0xff"},
                "inners": [{"dt": formatted, "i": "0xff"}],
                "i": "0xff",
            }
        )
        == obj
    )
    assert (
        DataClass.from_dict(
            {
                "dt": iso,
                "inner": {"dt": iso, "i": "0xff"},
                "inners": [{"dt": iso, "i": "0xff"}],
                "i": "0xff",
            },
            dialect=ISODialect,
        )
        == obj
    )


def test_debug_true_option_with_dialect(mocker):
    mocked_print = mocker.patch("builtins.print")

    @dataclass
    class DataClass(DataClassDictMixin):
        dt: date

        class Config(BaseConfig):
            debug = True
            code_generation_options = [ADD_DIALECT_SUPPORT]

    DataClass(date.today()).to_dict(dialect=FormattedDialect)
    mocked_print.assert_called()
    assert mocked_print.call_count == 6


def test_dialect_with_named_tuple_with_dialect_support():
    dt = date.today()
    ordinal = dt.toordinal()
    iso = dt.isoformat()
    obj = DataClassWithNamedTupleWithDialectSupport(
        x=MyNamedTuple(
            x=DataClassWithDialectSupport(dt, 255),
            y=DataClassWithoutDialects(dt, 255),
        )
    )
    dumped = {"x": [{"dt": ordinal, "i": "0xff"}, {"dt": iso, "i": 255}]}
    assert obj.to_dict(dialect=OrdinalDialect) == dumped
    assert (
        DataClassWithNamedTupleWithDialectSupport.from_dict(
            dumped, dialect=OrdinalDialect
        )
        == obj
    )


def test_dialect_with_named_tuple_without_dialect_support():
    dt = date.today()
    iso = dt.isoformat()
    obj = DataClassWithNamedTupleWithoutDialectSupport(
        x=MyNamedTuple(
            x=DataClassWithDialectSupport(dt, 255),
            y=DataClassWithoutDialects(dt, 255),
        )
    )
    dumped = {"x": [{"dt": iso, "i": 255}, {"dt": iso, "i": 255}]}
    assert obj.to_dict() == dumped
    assert (
        DataClassWithNamedTupleWithoutDialectSupport.from_dict(dumped) == obj
    )


def test_dialect_with_typed_dict_with_dialect_support():
    dt = date.today()
    ordinal = dt.toordinal()
    iso = dt.isoformat()
    obj = DataClassWithTypedDictWithDialectSupport(
        x=MyTypedDict(
            x=DataClassWithDialectSupport(dt, 255),
            y=DataClassWithoutDialects(dt, 255),
        )
    )
    dumped = {
        "x": {"x": {"dt": ordinal, "i": "0xff"}, "y": {"dt": iso, "i": 255}}
    }
    assert obj.to_dict(dialect=OrdinalDialect) == dumped
    assert (
        DataClassWithTypedDictWithDialectSupport.from_dict(
            dumped, dialect=OrdinalDialect
        )
        == obj
    )


def test_dialect_with_typed_dict_without_dialect_support():
    dt = date.today()
    iso = dt.isoformat()
    obj = DataClassWithTypedDictWithoutDialectSupport(
        x=MyTypedDict(
            x=DataClassWithDialectSupport(dt, 255),
            y=DataClassWithoutDialects(dt, 255),
        )
    )
    dumped = {"x": {"x": {"dt": iso, "i": 255}, "y": {"dt": iso, "i": 255}}}
    assert obj.to_dict() == dumped
    assert DataClassWithTypedDictWithoutDialectSupport.from_dict(dumped) == obj


def test_dialect_with_union_with_dialect_support():
    dt = date.today()
    ordinal = dt.toordinal()
    iso = dt.isoformat()
    obj = DataClassWithUnionWithDialectSupport(
        x=[
            DataClassWithDialectSupport(dt, 255),
            DataClassWithoutDialects(dt, 255),
        ]
    )
    dumped = {"x": [{"dt": ordinal, "i": "0xff"}, {"dt": iso, "i": 255}]}
    assert obj.to_dict(dialect=OrdinalDialect) == dumped
    assert (
        DataClassWithUnionWithDialectSupport.from_dict(
            dumped, dialect=OrdinalDialect
        )
        == obj
    )


def test_dialect_with_inheritance():
    dt = date.today()
    formatted = dt.strftime("%Y/%m/%d")
    entity1 = Entity1(dt)
    entity2 = Entity2(dt)

    assert (
        Entity1.from_dict({"dt1": formatted}, dialect=FormattedDialect)
        == entity1
    )
    assert (
        Entity2.from_dict({"dt2": formatted}, dialect=FormattedDialect)
        == entity2
    )
    assert entity1.to_dict(dialect=FormattedDialect) == {"dt1": formatted}
    assert entity2.to_dict(dialect=FormattedDialect) == {"dt2": formatted}


def test_msgpack_dialect_class_with_dependency_without_dialect():
    dt = date(2022, 6, 8)
    obj = MessagePackDataClass(
        b_1=b"123",
        b_2=bytearray([4, 5, 6]),
        dep_1=DataClassWithoutDialects(dt, 123),
        dep_2=GenericDataClassWithoutDialects(dt, 123),
    )
    d = {
        "b_1": b"123",
        "b_2": bytearray([4, 5, 6]),
        "dep_1": {"dt": "2022-06-08", "i": 123},
        "dep_2": {"dt": "2022-06-08", "i": 123},
    }
    encoded = msgpack_encoder(d)
    assert obj.to_msgpack() == encoded
    assert MessagePackDataClass.from_msgpack(encoded) == obj


def test_dataclass_omit_none_dialects():
    assert DataClassWithOptionalAndOmitNoneDialect().to_dict() == {}
    assert (
        DataClassWithOptionalAndOmitNoneDialectAndOmitNoneFalse().to_dict()
        == {}
    )
    assert (
        DataClassWithOptionalAndNotOmitNoneDialectAndOmitNoneTrue().to_dict()
        == {"x": None}
    )
    assert DataClassWithOptionalAndEmptyDialect().to_dict() == {}
    assert DataClassWithOptionalAndDialectSupport().to_dict() == {"x": None}
    assert (
        DataClassWithOptionalAndDialectSupport().to_dict(
            dialect=OmitNoneDialect
        )
        == {}
    )
    assert DataClassWithOptionalAndDialectSupport().to_dict(
        dialect=NotOmitNoneDialect
    ) == {"x": None}
    assert DataClassWithOptionalAndDialectSupport().to_dict(
        dialect=EmptyDialect
    ) == {"x": None}


def test_dataclass_omit_default_dialects():
    complete_dict = {
        "a": 42,
        "b": 3.14,
        "c": "foo",
        "d": False,
        "e": None,
        "f": None,
        "g": [1],
        "h": [{"dt": "2022-01-01", "i": 999}, {"dt": "2022-01-01", "i": 999}],
        "i": [],
        "j": [1, 2, 3],
        "k": [],
        "l": [1, 2, 3],
        "m": [],
        "n": [1, 2, 3],
        "o": MyStrEnum.VAL1,
    }
    assert DataClassWithDefaultAndOmitDefaultDialect().to_dict() == {}
    assert (
        DataClassWithDefaultAndOmitDefaultDialectAndOmitDefaultFalse().to_dict()
        == {}
    )
    assert (
        DataClassWithDefaultAndNotOmitDefaultDialectAndOmitDefaultTrue().to_dict()
        == complete_dict
    )
    assert DataClassWithDefaultAndEmptyDialect().to_dict() == {}
    assert DataClassWithDefaultAndDialectSupport().to_dict() == complete_dict
    assert (
        DataClassWithDefaultAndDialectSupport().to_dict(
            dialect=OmitDefaultDialect
        )
        == {}
    )
    assert (
        DataClassWithDefaultAndDialectSupport().to_dict(
            dialect=NotOmitDefaultDialect
        )
        == complete_dict
    )
    assert (
        DataClassWithDefaultAndDialectSupport().to_dict(dialect=EmptyDialect)
        == complete_dict
    )


def test_dialect_no_copy():
    class NoCopyDialect(Dialect):
        no_copy_collections = (list, dict, set)
        serialization_strategy = {int: {"serialize": pass_through}}

    @dataclass
    class DataClass(DataClassDictMixin):
        a: List[str]
        b: Set[str]
        c: typing.ChainMap[str, str]
        d: typing.OrderedDict[str, str]
        e: typing.Counter[str]
        f: typing.Dict[str, str]
        g: typing.Sequence[str]

        class Config(BaseConfig):
            dialect = NoCopyDialect

    obj = DataClass(
        a=["foo"],
        b={"foo"},
        c=collections.ChainMap({"foo": "bar"}),
        d=collections.OrderedDict({"foo": "bar"}),
        e=collections.Counter({"foo": 1}),
        f={"foo": "bar"},
        g=["foo"],
    )
    data = obj.to_dict()
    assert data["a"] is obj.a
    assert data["b"] is obj.b
    assert data["c"] is not obj.c
    assert data["d"] is not obj.d
    assert data["e"] is not obj.e
    assert data["f"] is obj.f
    assert data["g"] is not obj.g


def test_dialect_merge():
    class DialectA(Dialect):
        omit_none = True
        omit_default = True
        no_copy_collections = [set, dict]
        serialization_strategy = {
            date: {
                "serialize": date.toordinal,
                "deserialize": date.fromordinal,
            },
            int: HexSerializationStrategy(),
            float: {
                "serialize": pass_through,
                "deserialize": float,
            },
        }

    class DialectB(Dialect):
        omit_none = False
        omit_default = False
        no_copy_collections = [list]
        serialization_strategy = {
            date: pass_through,
            int: {
                "serialize": int,
                "deserialize": int,
            },
            float: {
                "serialize": float,
            },
        }

    DialectC = DialectA.merge(DialectB)
    assert DialectC.omit_none is False
    assert DialectC.omit_default is False
    assert DialectC.no_copy_collections == [list]
    assert DialectC.serialization_strategy == {
        date: pass_through,
        int: {
            "serialize": int,
            "deserialize": int,
        },
        float: {
            "serialize": float,
            "deserialize": float,
        },
    }



from dataclasses import dataclass, field
from datetime import date, datetime

import pytest

from mashumaro import DataClassDictMixin, field_options
from mashumaro.helper import pass_through
from mashumaro.types import SerializationStrategy


def test_field_options_helper():
    assert field_options() == {
        "serialize": None,
        "deserialize": None,
        "serialization_strategy": None,
        "alias": None,
    }

    def serialize(x):
        return x  # pragma: no cover

    def deserialize(x):
        return x  # pragma: no cover

    class TestSerializationStrategy(SerializationStrategy):  # pragma: no cover
        def deserialize(self, value):
            return value

        def serialize(self, value):
            return value

    serialization_strategy = TestSerializationStrategy()
    alias = "alias"

    assert field_options(
        serialize=serialize,
        deserialize=deserialize,
        serialization_strategy=serialization_strategy,
        alias=alias,
    ) == {
        "serialize": serialize,
        "deserialize": deserialize,
        "serialization_strategy": serialization_strategy,
        "alias": alias,
    }


def test_pass_through():
    with pytest.raises(NotImplementedError):
        pass_through()
    assert pass_through.serialize(123) == 123
    assert pass_through.deserialize(123) == 123


def test_dataclass_with_pass_through():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: datetime = field(
            metadata=field_options(
                serialize=pass_through,
                deserialize=pass_through,
            )
        )
        y: date = field(
            metadata=field_options(serialization_strategy=pass_through)
        )

    x = datetime.now()
    y = x.date()
    instance = DataClass(x, y)
    assert instance.to_dict() == {"x": x, "y": y}
    assert instance.from_dict({"x": x, "y": y}) == instance



import json
from dataclasses import dataclass

from typing_extensions import Literal, TypeVarTuple

from mashumaro.config import BaseConfig
from mashumaro.jsonschema import DRAFT_2020_12, OPEN_API_3_1
from mashumaro.jsonschema.builder import JSONSchemaBuilder, build_json_schema
from mashumaro.jsonschema.schema import Instance

Ts = TypeVarTuple("Ts")


@dataclass
class A:
    a: int


@dataclass
class B:
    b: float


def test_instance():
    @dataclass
    class DataClass:
        class Config(BaseConfig):
            pass

    instance = Instance(int)
    assert instance.metadata == {}
    assert instance.owner_class is None
    assert instance.metadata == {}
    assert instance.get_self_config() is BaseConfig

    instance = Instance(DataClass)
    assert instance.owner_class is None
    assert instance.metadata == {}
    assert instance.get_self_config() is DataClass.Config

    derived_instance = instance.derive(type=int)
    assert derived_instance.owner_class is DataClass
    assert derived_instance.metadata == {}
    assert derived_instance.get_self_config() is BaseConfig


def test_jsonschema_json_simple():
    assert json.loads(build_json_schema(int).to_json()) == {"type": "integer"}


def test_jsonschema_json_literal_none():
    @dataclass
    class DataClass:
        x: Literal[None] = None

    assert json.loads(build_json_schema(DataClass).to_json()) == {
        "type": "object",
        "title": "DataClass",
        "properties": {"x": {"const": None, "default": None}},
        "additionalProperties": False,
    }


def test_jsonschema_builder_draft_2020_12():
    builder = JSONSchemaBuilder(dialect=DRAFT_2020_12)
    assert builder.build(A).to_dict() == {
        "type": "object",
        "title": "A",
        "properties": {"a": {"type": "integer"}},
        "additionalProperties": False,
        "required": ["a"],
    }
    assert builder.get_definitions().to_dict() == {}


def test_jsonschema_builder_draft_2020_12_with_refs():
    builder = JSONSchemaBuilder(dialect=DRAFT_2020_12, all_refs=True)
    assert builder.build(A).to_dict() == {"$ref": "#/$defs/A"}
    assert builder.build(B).to_dict() == {"$ref": "#/$defs/B"}
    assert builder.get_definitions().to_dict() == {
        "A": {
            "type": "object",
            "title": "A",
            "properties": {"a": {"type": "integer"}},
            "additionalProperties": False,
            "required": ["a"],
        },
        "B": {
            "type": "object",
            "title": "B",
            "properties": {"b": {"type": "number"}},
            "additionalProperties": False,
            "required": ["b"],
        },
    }


def test_jsonschema_builder_open_api_3_1():
    builder = JSONSchemaBuilder(dialect=OPEN_API_3_1)
    assert builder.build(A).to_dict() == {"$ref": "#/components/schemas/A"}
    assert builder.build(B).to_dict() == {"$ref": "#/components/schemas/B"}
    assert builder.get_definitions().to_dict() == {
        "A": {
            "type": "object",
            "title": "A",
            "properties": {"a": {"type": "integer"}},
            "additionalProperties": False,
            "required": ["a"],
        },
        "B": {
            "type": "object",
            "title": "B",
            "properties": {"b": {"type": "number"}},
            "additionalProperties": False,
            "required": ["b"],
        },
    }


def test_jsonschema_builder_open_api_3_1_without_refs():
    builder = JSONSchemaBuilder(dialect=OPEN_API_3_1, all_refs=False)
    assert builder.build(A).to_dict() == {
        "type": "object",
        "title": "A",
        "properties": {"a": {"type": "integer"}},
        "additionalProperties": False,
        "required": ["a"],
    }
    assert builder.build(B).to_dict() == {
        "type": "object",
        "title": "B",
        "properties": {"b": {"type": "number"}},
        "additionalProperties": False,
        "required": ["b"],
    }
    assert builder.get_definitions().to_dict() == {}



from datetime import date
from typing import Dict, List

import tomli_w

from mashumaro.codecs.toml import (
    TOMLDecoder,
    TOMLEncoder,
    toml_decode,
    toml_encode,
)
from mashumaro.dialect import Dialect


class MyDialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": date.toordinal,
            "deserialize": date.fromordinal,
        },
    }


def test_toml_decode():
    data = tomli_w.dumps({"x": [date(2023, 9, 22), date(2023, 9, 23)]})
    assert toml_decode(data, Dict[str, List[date]]) == {
        "x": [
            date(2023, 9, 22),
            date(2023, 9, 23),
        ]
    }


def test_toml_encode():
    data = tomli_w.dumps({"x": [date(2023, 9, 22), date(2023, 9, 23)]})
    assert (
        toml_encode(
            {"x": [date(2023, 9, 22), date(2023, 9, 23)]},
            Dict[str, List[date]],
        )
        == data
    )


def test_decoder_with_default_dialect():
    data = tomli_w.dumps({"x": [738785, 738786]})
    decoder = TOMLDecoder(Dict[str, List[date]], default_dialect=MyDialect)
    assert decoder.decode(data) == {
        "x": [
            date(2023, 9, 22),
            date(2023, 9, 23),
        ]
    }


def test_encoder_with_default_dialect():
    data = tomli_w.dumps({"x": [738785, 738786]})
    encoder = TOMLEncoder(Dict[str, List[date]], default_dialect=MyDialect)
    assert (
        encoder.encode({"x": [date(2023, 9, 22), date(2023, 9, 23)]}) == data
    )



from dataclasses import dataclass
from datetime import date
from enum import Enum
from typing import ClassVar, Tuple, Union

import pytest
from typing_extensions import Annotated, Final, Literal

from mashumaro import DataClassDictMixin
from mashumaro.codecs import BasicDecoder
from mashumaro.exceptions import InvalidFieldValue
from mashumaro.types import Discriminator

DT_STR = "2023-05-30"
DT_DATE = date(2023, 5, 30)

X_STR = {"x": "2023-05-30", "type": "str"}
X_DATE = {"x": "2023-05-30", "type": "date"}
X_DATE_SUBTYPE = {"x": "2023-05-30", "type": "date_subtype"}


class VariantType(str, Enum):
    STR = "str"
    DATE = "date"
    DATE_SUBTYPE = "date_subtype"


@dataclass
class UnannotatedVariantStr:
    x: str
    type = "str"


@dataclass
class ClassVarVariantStr:
    x: str
    type: ClassVar[str] = "str"


@dataclass
class FinalVariantStr:
    x: str
    type: Final[str] = "str"


@dataclass
class LiteralVariantStr:
    x: str
    type: Literal["str"] = "str"


@dataclass
class EnumVariantStr:
    x: str
    type: VariantType = VariantType.STR


@dataclass
class UnannotatedVariantDate:
    x: date
    type = "date"


@dataclass
class ClassVarVariantDate:
    x: date
    type: ClassVar[str] = "date"


@dataclass
class FinalVariantDate:
    x: date
    type: Final[str] = "date"


@dataclass
class LiteralVariantDate:
    x: date
    type: Literal["date"] = "date"


@dataclass
class EnumVariantDate:
    x: date
    type: VariantType = VariantType.DATE


@dataclass
class UnannotatedVariantDateSubtype(UnannotatedVariantDate):
    x: date
    type = "date_subtype"


@dataclass
class ClassVarVariantDateSubtype(ClassVarVariantDate):
    x: date
    type: ClassVar[str] = "date_subtype"


@dataclass
class FinalVariantDateSubtype(FinalVariantDate):
    x: date
    type: Final[str] = "date_subtype"


@dataclass
class LiteralVariantDateSubtype(LiteralVariantDate):
    x: date
    type: Literal["date_subtype"] = "date_subtype"


@dataclass
class EnumVariantDateSubtype(EnumVariantDate):
    x: date
    type: VariantType = VariantType.DATE_SUBTYPE


@dataclass
class _ByFieldWithSupertypes:
    unannotated: Annotated[
        Union[UnannotatedVariantStr, UnannotatedVariantDate],
        Discriminator("type", include_supertypes=True),
    ]
    class_var: Annotated[
        Union[ClassVarVariantStr, ClassVarVariantDate],
        Discriminator("type", include_supertypes=True),
    ]
    final: Annotated[
        Union[FinalVariantStr, FinalVariantDate],
        Discriminator("type", include_supertypes=True),
    ]
    literal: Annotated[
        Union[LiteralVariantStr, LiteralVariantDate],
        Discriminator("type", include_supertypes=True),
    ]
    enum: Annotated[
        Union[EnumVariantStr, EnumVariantDate],
        Discriminator("type", include_supertypes=True),
    ]


@dataclass
class ByFieldWithSupertypes(_ByFieldWithSupertypes, DataClassDictMixin):
    pass


@dataclass
class _ByFieldWithSubtypes:
    unannotated: Annotated[
        Union[UnannotatedVariantStr, UnannotatedVariantDate],
        Discriminator("type", include_subtypes=True),
    ]
    class_var: Annotated[
        Union[ClassVarVariantStr, ClassVarVariantDate],
        Discriminator("type", include_subtypes=True),
    ]
    final: Annotated[
        Union[FinalVariantStr, FinalVariantDate],
        Discriminator("type", include_subtypes=True),
    ]
    literal: Annotated[
        Union[LiteralVariantStr, LiteralVariantDate],
        Discriminator("type", include_subtypes=True),
    ]
    enum: Annotated[
        Union[EnumVariantStr, EnumVariantDate],
        Discriminator("type", include_subtypes=True),
    ]


@dataclass
class ByFieldWithSubtypes(_ByFieldWithSubtypes, DataClassDictMixin):
    pass


@dataclass
class _BySupertypes:
    unannotated: Annotated[
        Union[UnannotatedVariantStr, UnannotatedVariantDate],
        Discriminator(include_supertypes=True),
    ]
    class_var: Annotated[
        Union[ClassVarVariantStr, ClassVarVariantDate],
        Discriminator(include_supertypes=True),
    ]
    final: Annotated[
        Union[FinalVariantStr, FinalVariantDate],
        Discriminator(include_supertypes=True),
    ]
    literal: Annotated[
        Union[LiteralVariantStr, LiteralVariantDate],
        Discriminator(include_supertypes=True),
    ]
    enum: Annotated[
        Union[EnumVariantStr, EnumVariantDate],
        Discriminator(include_supertypes=True),
    ]


@dataclass
class BySupertypes(_BySupertypes, DataClassDictMixin):
    pass


@dataclass
class _BySubtypes:
    unannotated: Annotated[
        Union[UnannotatedVariantStr, UnannotatedVariantDate],
        Discriminator(include_subtypes=True),
    ]
    class_var: Annotated[
        Union[ClassVarVariantStr, ClassVarVariantDate],
        Discriminator(include_subtypes=True),
    ]
    final: Annotated[
        Union[FinalVariantStr, FinalVariantDate],
        Discriminator(include_subtypes=True),
    ]
    literal: Annotated[
        Union[LiteralVariantStr, LiteralVariantDate],
        Discriminator(include_subtypes=True),
    ]
    enum: Annotated[
        Union[EnumVariantStr, EnumVariantDate],
        Discriminator(include_subtypes=True),
    ]


@dataclass
class BySubtypes(_BySubtypes, DataClassDictMixin):
    pass


@dataclass
class _BySupertypesAndSubtypes:
    unannotated: Annotated[
        Union[UnannotatedVariantStr, UnannotatedVariantDate],
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]
    class_var: Annotated[
        Union[ClassVarVariantStr, ClassVarVariantDate],
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]
    final: Annotated[
        Union[FinalVariantStr, FinalVariantDate],
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]
    literal: Annotated[
        Union[LiteralVariantStr, LiteralVariantDate],
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]
    enum: Annotated[
        Union[EnumVariantStr, EnumVariantDate],
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]


@dataclass
class BySupertypesAndSubtypes(_BySupertypesAndSubtypes, DataClassDictMixin):
    pass


@dataclass
class _ByFieldWithSupertypesAndSubtypes:
    unannotated: Annotated[
        Union[UnannotatedVariantStr, UnannotatedVariantDate],
        Discriminator("type", include_supertypes=True, include_subtypes=True),
    ]
    class_var: Annotated[
        Union[ClassVarVariantStr, ClassVarVariantDate],
        Discriminator("type", include_supertypes=True, include_subtypes=True),
    ]
    final: Annotated[
        Union[FinalVariantStr, FinalVariantDate],
        Discriminator("type", include_supertypes=True, include_subtypes=True),
    ]
    literal: Annotated[
        Union[LiteralVariantStr, LiteralVariantDate],
        Discriminator("type", include_supertypes=True, include_subtypes=True),
    ]
    enum: Annotated[
        Union[EnumVariantStr, EnumVariantDate],
        Discriminator("type", include_supertypes=True, include_subtypes=True),
    ]


@dataclass
class ByFieldWithSupertypesAndSubtypes(
    _ByFieldWithSupertypesAndSubtypes, DataClassDictMixin
):
    pass


@dataclass
class _ByFieldAndByFieldWithSubtypesInOneField:
    x: Tuple[
        Annotated[
            Union[UnannotatedVariantStr, UnannotatedVariantDate],
            Discriminator("type", include_supertypes=True),
        ],
        Annotated[
            Union[UnannotatedVariantStr, UnannotatedVariantDate],
            Discriminator("type", include_subtypes=True),
        ],
    ]


@dataclass
class ByFieldAndByFieldWithSubtypesInOneField(
    _ByFieldAndByFieldWithSubtypesInOneField, DataClassDictMixin
):
    pass


@dataclass
class VariantWitCustomTagger1:
    pass


@dataclass
class VariantWitCustomTagger2:
    pass


@dataclass
class _VariantWitCustomTaggerOwner:
    x: Annotated[
        Union[VariantWitCustomTagger1, VariantWitCustomTagger2],
        Discriminator(
            field="type",
            include_supertypes=True,
            variant_tagger_fn=lambda cls: cls.__name__.lower(),
        ),
    ]


@dataclass
class VariantWitCustomTaggerOwner(
    _VariantWitCustomTaggerOwner, DataClassDictMixin
):
    pass


@dataclass
class _VariantWitCustomTaggerWithMultipleTagsOwner:
    x: Annotated[
        Union[VariantWitCustomTagger1, VariantWitCustomTagger2],
        Discriminator(
            field="type",
            include_supertypes=True,
            variant_tagger_fn=lambda cls: [
                cls.__name__.lower(),
                cls.__name__.upper(),
            ],
        ),
    ]


@dataclass
class VariantWitCustomTaggerWithMultipleTagsOwner(
    _VariantWitCustomTaggerWithMultipleTagsOwner, DataClassDictMixin
):
    pass


def test_by_field_with_supertypes():
    decoder = BasicDecoder(_ByFieldWithSupertypes)

    for func, cls in (
        (ByFieldWithSupertypes.from_dict, ByFieldWithSupertypes),
        (decoder.decode, _ByFieldWithSupertypes),
    ):
        assert func(
            {
                "unannotated": X_STR,
                "class_var": X_STR,
                "literal": X_STR,
                "final": X_STR,
                "enum": X_STR,
            }
        ) == cls(
            unannotated=UnannotatedVariantStr(DT_STR),
            class_var=ClassVarVariantStr(DT_STR),
            literal=LiteralVariantStr(DT_STR),
            final=FinalVariantStr(DT_STR),
            enum=EnumVariantStr(DT_STR),
        )

    for func, cls in (
        (ByFieldWithSupertypes.from_dict, ByFieldWithSupertypes),
        (decoder.decode, _ByFieldWithSupertypes),
    ):
        assert func(
            {
                "unannotated": X_DATE,
                "class_var": X_DATE,
                "literal": X_DATE,
                "final": X_DATE,
                "enum": X_DATE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDate(DT_DATE),
            class_var=ClassVarVariantDate(DT_DATE),
            literal=LiteralVariantDate(DT_DATE),
            final=FinalVariantDate(DT_DATE),
            enum=EnumVariantDate(DT_DATE),
        )

    for func in (ByFieldWithSupertypes.from_dict, decoder.decode):
        with pytest.raises(InvalidFieldValue) as exc_info:
            func({"unannotated": {"x": "2023-05-30", "type": "date_subtype"}})
        assert exc_info.value.field_name == "unannotated"


def test_by_field_with_subtypes():
    decoder = BasicDecoder(_ByFieldWithSubtypes)

    for func, cls in (
        (ByFieldWithSubtypes.from_dict, ByFieldWithSubtypes),
        (decoder.decode, _ByFieldWithSubtypes),
    ):
        assert func(
            {
                "unannotated": X_DATE_SUBTYPE,
                "class_var": X_DATE_SUBTYPE,
                "literal": X_DATE_SUBTYPE,
                "final": X_DATE_SUBTYPE,
                "enum": X_DATE_SUBTYPE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDateSubtype(DT_DATE),
            class_var=ClassVarVariantDateSubtype(DT_DATE),
            literal=LiteralVariantDateSubtype(DT_DATE),
            final=FinalVariantDateSubtype(DT_DATE),
            enum=EnumVariantDateSubtype(DT_DATE),
        )

    for func in (ByFieldWithSubtypes.from_dict, decoder.decode):
        with pytest.raises(InvalidFieldValue) as exc_info:
            func(
                {
                    "unannotated": X_STR,
                    "class_var": X_STR,
                    "literal": X_STR,
                    "final": X_STR,
                    "enum": X_STR,
                }
            )
        assert exc_info.value.field_name == "unannotated"

    for func in (ByFieldWithSubtypes.from_dict, decoder.decode):
        with pytest.raises(InvalidFieldValue) as exc_info:
            func(
                {
                    "unannotated": X_DATE,
                    "class_var": X_DATE,
                    "literal": X_DATE,
                    "final": X_DATE,
                    "enum": X_DATE,
                }
            )
        assert exc_info.value.field_name == "unannotated"


def test_by_field_with_supertypes_and_subtypes():
    decoder = BasicDecoder(_ByFieldWithSupertypesAndSubtypes)

    for func, cls in (
        (
            ByFieldWithSupertypesAndSubtypes.from_dict,
            ByFieldWithSupertypesAndSubtypes,
        ),
        (decoder.decode, _ByFieldWithSupertypesAndSubtypes),
    ):
        assert func(
            {
                "unannotated": X_STR,
                "class_var": X_STR,
                "literal": X_STR,
                "final": X_STR,
                "enum": X_STR,
            }
        ) == cls(
            unannotated=UnannotatedVariantStr(DT_STR),
            class_var=ClassVarVariantStr(DT_STR),
            literal=LiteralVariantStr(DT_STR),
            final=FinalVariantStr(DT_STR),
            enum=EnumVariantStr(DT_STR),
        )

    for func, cls in (
        (
            ByFieldWithSupertypesAndSubtypes.from_dict,
            ByFieldWithSupertypesAndSubtypes,
        ),
        (decoder.decode, _ByFieldWithSupertypesAndSubtypes),
    ):
        assert func(
            {
                "unannotated": X_DATE,
                "class_var": X_DATE,
                "literal": X_DATE,
                "final": X_DATE,
                "enum": X_DATE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDate(DT_DATE),
            class_var=ClassVarVariantDate(DT_DATE),
            literal=LiteralVariantDate(DT_DATE),
            final=FinalVariantDate(DT_DATE),
            enum=EnumVariantDate(DT_DATE),
        )

    for func, cls in (
        (
            ByFieldWithSupertypesAndSubtypes.from_dict,
            ByFieldWithSupertypesAndSubtypes,
        ),
        (decoder.decode, _ByFieldWithSupertypesAndSubtypes),
    ):
        assert func(
            {
                "unannotated": X_DATE_SUBTYPE,
                "class_var": X_DATE_SUBTYPE,
                "literal": X_DATE_SUBTYPE,
                "final": X_DATE_SUBTYPE,
                "enum": X_DATE_SUBTYPE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDateSubtype(DT_DATE),
            class_var=ClassVarVariantDateSubtype(DT_DATE),
            literal=LiteralVariantDateSubtype(DT_DATE),
            final=FinalVariantDateSubtype(DT_DATE),
            enum=EnumVariantDateSubtype(DT_DATE),
        )


def test_by_supertypes():
    decoder = BasicDecoder(_BySupertypes)

    for func, cls in (
        (BySupertypes.from_dict, BySupertypes),
        (decoder.decode, _BySupertypes),
    ):
        assert func(
            {
                "unannotated": X_STR,
                "class_var": X_STR,
                "literal": X_STR,
                "final": X_STR,
                "enum": X_STR,
            }
        ) == cls(
            unannotated=UnannotatedVariantStr(DT_STR),
            class_var=ClassVarVariantStr(DT_STR),
            literal=LiteralVariantStr(DT_STR),
            final=FinalVariantStr(DT_STR),
            enum=EnumVariantStr(DT_STR),
        )

    for func, cls in (
        (BySupertypes.from_dict, BySupertypes),
        (decoder.decode, _BySupertypes),
    ):
        assert func(
            {
                "unannotated": X_DATE,
                "class_var": X_DATE,
                "literal": X_DATE,
                "final": X_DATE,
                "enum": X_DATE,
            }
        ) == cls(
            unannotated=UnannotatedVariantStr(DT_STR),
            class_var=ClassVarVariantStr(DT_STR),
            literal=LiteralVariantDate(DT_DATE),
            # final without field discriminator can lead to unexpected result
            final=FinalVariantStr(DT_STR, type=VariantType.DATE),
            # enum without field discriminator can lead to unexpected result
            enum=EnumVariantStr(DT_STR, type=VariantType.DATE),
        )

    for func in (BySupertypes.from_dict, decoder.decode):
        with pytest.raises(InvalidFieldValue) as exc_info:
            func(
                {
                    "unannotated": X_DATE_SUBTYPE,
                    "class_var": X_DATE_SUBTYPE,
                    "literal": X_DATE_SUBTYPE,
                    "final": X_DATE_SUBTYPE,
                    "enum": X_DATE_SUBTYPE,
                }
            )
        assert exc_info.value.field_name == "literal"


def test_by_subtypes():
    decoder = BasicDecoder(_BySubtypes)

    for func, cls in (
        (BySubtypes.from_dict, BySubtypes),
        (decoder.decode, _BySubtypes),
    ):
        assert func(
            {
                "unannotated": X_DATE_SUBTYPE,
                "class_var": X_DATE_SUBTYPE,
                "literal": X_DATE_SUBTYPE,
                "final": X_DATE_SUBTYPE,
                "enum": X_DATE_SUBTYPE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDateSubtype(DT_DATE),
            class_var=ClassVarVariantDateSubtype(DT_DATE),
            literal=LiteralVariantDateSubtype(DT_DATE),
            final=FinalVariantDateSubtype(DT_DATE),
            enum=EnumVariantDateSubtype(DT_DATE),
        )

    for func, cls in (
        (BySubtypes.from_dict, BySubtypes),
        (decoder.decode, _BySubtypes),
    ):
        assert func(
            {
                "unannotated": X_DATE,
                "class_var": X_DATE,
                "literal": X_DATE_SUBTYPE,
                "final": X_DATE,
                "enum": X_DATE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDateSubtype(DT_DATE),
            class_var=ClassVarVariantDateSubtype(DT_DATE),
            literal=LiteralVariantDateSubtype(DT_DATE),
            # final without field discriminator can lead to unexpected result
            final=FinalVariantDateSubtype(DT_DATE, type=VariantType.DATE),
            # enum without field discriminator can lead to unexpected result
            enum=EnumVariantDateSubtype(DT_DATE, type=VariantType.DATE),
        )

    for func in (BySubtypes.from_dict, decoder.decode):
        with pytest.raises(InvalidFieldValue) as exc_info:
            func(
                {
                    "unannotated": X_STR,
                    "class_var": X_STR,
                    "literal": X_STR,
                    "final": X_STR,
                    "enum": X_STR,
                }
            )
        assert exc_info.value.field_name == "literal"


def test_by_supertypes_and_subtypes():
    decoder = BasicDecoder(_BySupertypesAndSubtypes)

    for func, cls in (
        (BySupertypesAndSubtypes.from_dict, BySupertypesAndSubtypes),
        (decoder.decode, _BySupertypesAndSubtypes),
    ):
        assert func(
            {
                "unannotated": X_DATE_SUBTYPE,
                "class_var": X_DATE_SUBTYPE,
                "literal": X_DATE_SUBTYPE,
                "final": X_DATE_SUBTYPE,
                "enum": X_DATE_SUBTYPE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDateSubtype(DT_DATE),
            class_var=ClassVarVariantDateSubtype(DT_DATE),
            literal=LiteralVariantDateSubtype(DT_DATE),
            final=FinalVariantDateSubtype(DT_DATE),
            enum=EnumVariantDateSubtype(DT_DATE),
        )

    for func, cls in (
        (BySupertypesAndSubtypes.from_dict, BySupertypesAndSubtypes),
        (decoder.decode, _BySupertypesAndSubtypes),
    ):
        assert func(
            {
                "unannotated": X_STR,
                "class_var": X_STR,
                "literal": X_STR,
                "final": X_STR,
                "enum": X_STR,
            }
        ) == cls(
            unannotated=UnannotatedVariantDateSubtype(DT_DATE),
            class_var=ClassVarVariantDateSubtype(DT_DATE),
            literal=LiteralVariantStr(DT_STR),
            # using final without field discriminator
            # can lead to unexpected result
            final=FinalVariantDateSubtype(DT_DATE, type=VariantType.STR),
            # using enum without field discriminator
            # can lead to unexpected result
            enum=EnumVariantDateSubtype(DT_DATE, type=VariantType.STR),
        )

    for func, cls in (
        (BySupertypesAndSubtypes.from_dict, BySupertypesAndSubtypes),
        (decoder.decode, _BySupertypesAndSubtypes),
    ):
        assert func(
            {
                "unannotated": X_DATE,
                "class_var": X_DATE,
                "literal": X_DATE,
                "final": X_DATE,
                "enum": X_DATE,
            }
        ) == cls(
            unannotated=UnannotatedVariantDateSubtype(DT_DATE),
            class_var=ClassVarVariantDateSubtype(DT_DATE),
            literal=LiteralVariantDate(DT_DATE),
            # final without field discriminator can lead to unexpected result
            final=FinalVariantDateSubtype(DT_DATE, type=VariantType.DATE),
            # enum without field discriminator can lead to unexpected result
            enum=EnumVariantDateSubtype(DT_DATE, type=VariantType.DATE),
        )


def test_tuple_with_discriminated_elements():
    decoder = BasicDecoder(_ByFieldAndByFieldWithSubtypesInOneField)

    for func, cls in (
        (
            ByFieldAndByFieldWithSubtypesInOneField.from_dict,
            ByFieldAndByFieldWithSubtypesInOneField,
        ),
        (decoder.decode, _ByFieldAndByFieldWithSubtypesInOneField),
    ):
        assert func({"x": [X_STR, X_DATE_SUBTYPE]}) == cls(
            (
                UnannotatedVariantStr(DT_STR),
                UnannotatedVariantDateSubtype(DT_DATE),
            ),
        )

    for func in (
        ByFieldAndByFieldWithSubtypesInOneField.from_dict,
        decoder.decode,
    ):
        with pytest.raises(InvalidFieldValue):
            func({"x": [X_DATE_SUBTYPE, X_DATE_SUBTYPE]})

    for func in (
        ByFieldAndByFieldWithSubtypesInOneField.from_dict,
        decoder.decode,
    ):
        with pytest.raises(InvalidFieldValue):
            func({"x": [X_STR, X_STR]})


def test_by_field_with_subtypes_with_custom_variant_tagger():
    decoder = BasicDecoder(_VariantWitCustomTaggerOwner)

    for func, cls in (
        (VariantWitCustomTaggerOwner.from_dict, VariantWitCustomTaggerOwner),
        (decoder.decode, _VariantWitCustomTaggerOwner),
    ):
        assert func({"x": {"type": "variantwitcustomtagger1"}}) == cls(
            VariantWitCustomTagger1()
        )
        assert func({"x": {"type": "variantwitcustomtagger2"}}) == cls(
            VariantWitCustomTagger2()
        )
        with pytest.raises(InvalidFieldValue):
            func({"x": {"type": "unknown"}})


def test_by_field_with_subtypes_with_custom_variant_tagger_and_multiple_tags():
    decoder = BasicDecoder(_VariantWitCustomTaggerWithMultipleTagsOwner)

    for func, cls in (
        (
            VariantWitCustomTaggerWithMultipleTagsOwner.from_dict,
            VariantWitCustomTaggerWithMultipleTagsOwner,
        ),
        (decoder.decode, _VariantWitCustomTaggerWithMultipleTagsOwner),
    ):
        assert func({"x": {"type": "variantwitcustomtagger1"}}) == cls(
            VariantWitCustomTagger1()
        )
        assert func({"x": {"type": "variantwitcustomtagger2"}}) == cls(
            VariantWitCustomTagger2()
        )
        assert func({"x": {"type": "VARIANTWITCUSTOMTAGGER1"}}) == cls(
            VariantWitCustomTagger1()
        )
        assert func({"x": {"type": "VARIANTWITCUSTOMTAGGER2"}}) == cls(
            VariantWitCustomTagger2()
        )
        with pytest.raises(InvalidFieldValue):
            func({"x": {"type": "unknown"}})



from dataclasses import dataclass
from typing import List, Union

import pytest

from mashumaro import DataClassDictMixin
from mashumaro.codecs import BasicDecoder
from mashumaro.core.meta.helpers import type_name
from mashumaro.exceptions import (
    ExtraKeysError,
    InvalidFieldValue,
    MissingDiscriminatorError,
    MissingField,
    SuitableVariantNotFoundError,
    ThirdPartyModuleNotFoundError,
    UnresolvedTypeReferenceError,
    UnserializableField,
    UnsupportedDeserializationEngine,
    UnsupportedSerializationEngine,
)


def test_missing_field_simple_field_type_name():
    exc = MissingField("x", int, object)
    assert exc.field_type_name == "int"


def test_missing_field_generic_field_type_name():
    exc = MissingField("x", List[int], object)
    assert exc.field_type_name == "List[int]"


def test_missing_field_holder_class_name():
    exc = MissingField("x", int, object)
    assert exc.holder_class_name == "object"
    exc = MissingField("x", int, List[int])
    assert exc.holder_class_name == "List[int]"


def test_missing_field_str():
    exc = MissingField("x", int, object)
    assert str(exc) == 'Field "x" of type int is missing in object instance'


def test_unserializable_field_simple_field_type_name():
    exc = UnserializableField("x", int, object)
    assert exc.field_type_name == "int"


def test_unserializable_field_generic_field_type_name():
    exc = UnserializableField("x", List[int], object)
    assert exc.field_type_name == "List[int]"


def test_unserializable_field_holder_class_name():
    exc = UnserializableField("x", int, object)
    assert exc.holder_class_name == "object"
    exc = UnserializableField("x", int, List[int])
    assert exc.holder_class_name == "List[int]"


def test_unserializable_field_str():
    exc = UnserializableField("x", int, object)
    assert str(exc) == 'Field "x" of type int in object is not serializable'


def test_unserializable_field_with_msg_str():
    exc = UnserializableField("x", int, object, "test message")
    assert (
        str(exc) == 'Field "x" of type int in object '
        "is not serializable: test message"
    )


def test_invalid_field_value_simple_field_type_name():
    exc = InvalidFieldValue("x", int, "y", object)
    assert exc.field_type_name == "int"


def test_invalid_field_value_generic_field_type_name():
    exc = InvalidFieldValue("x", List[int], "y", object)
    assert exc.field_type_name == "List[int]"


def test_invalid_field_value_holder_class_name():
    exc = InvalidFieldValue("x", int, "y", object)
    assert exc.holder_class_name == "object"
    exc = InvalidFieldValue("x", int, "y", List[int])
    assert exc.holder_class_name == "List[int]"


def test_invalid_field_value_str():
    exc = InvalidFieldValue("x", int, "y", object)
    assert (
        str(exc) == "Field \"x\" of type int in object has invalid value 'y'"
    )


def test_invalid_field_value_with_msg_str():
    exc = InvalidFieldValue("x", int, "y", object, "test message")
    assert (
        str(exc) == 'Field "x" of type int in object '
        "has invalid value 'y': test message"
    )


def test_third_party_module_not_found_error_holder_class_name():
    exc = ThirdPartyModuleNotFoundError("third_party", "x", object)
    assert exc.holder_class_name == "object"
    exc = ThirdPartyModuleNotFoundError("third_party", "x", List[int])
    assert exc.holder_class_name == "List[int]"


def test_third_party_module_not_found_error_str():
    exc = ThirdPartyModuleNotFoundError("third_party", "x", object)
    assert (
        str(exc) == 'Install "third_party" to use it as the serialization '
        'method for the field "x" in object'
    )


def test_unsupported_deserialization_engine():
    exc = UnsupportedDeserializationEngine("x", int, object, "engine_name")
    assert exc.field_type_name == "int"
    assert exc.holder_class_name == "object"
    assert (
        str(exc) == 'Field "x" of type int in object is not serializable: '
        'Unsupported deserialization engine "engine_name"'
    )


def test_unsupported_serialization_engine():
    exc = UnsupportedSerializationEngine("x", int, object, "engine_name")
    assert exc.field_type_name == "int"
    assert exc.holder_class_name == "object"
    assert (
        str(exc) == 'Field "x" of type int in object is not serializable: '
        'Unsupported serialization engine "engine_name"'
    )


def test_unresolved_type_reference_error():
    exc = UnresolvedTypeReferenceError(object, "x")
    assert exc.holder_class_name == "object"
    assert exc.name == "x"
    assert (
        str(exc) == "Class object has unresolved type reference "
        "x in some of its fields"
    )


def test_missing_discriminator_error():
    exc = MissingDiscriminatorError("x")
    assert exc.field_name == "x"
    assert str(exc) == "Discriminator 'x' is missing"


def test_suitable_variant_not_found_error():
    exc = SuitableVariantNotFoundError(Union[str, int], "type", 42)
    assert exc.discriminator_value == 42
    assert exc.variants_type == Union[str, int]
    assert exc.discriminator_name == "type"
    assert str(exc) == (
        "typing.Union[str, int] has no subtype with attribute 'type' "
        "equal to 42"
    )

    exc = SuitableVariantNotFoundError(Union[str, int], "type")
    assert exc.discriminator_value is None
    assert exc.variants_type == Union[str, int]
    assert exc.discriminator_name == "type"
    assert str(exc) == "typing.Union[str, int] has no suitable subtype"


def test_deserialize_dataclass_from_wrong_value_type():
    @dataclass
    class MyClass(DataClassDictMixin):
        x: str

    with pytest.raises(ValueError) as exc_info:
        MyClass.from_dict(42)
    assert str(exc_info.value) == (
        f"Argument for {type_name(MyClass)}."
        f"__mashumaro_from_dict__ method should be a dict instance"
    )

    decoder = BasicDecoder(MyClass)
    with pytest.raises(ValueError) as exc_info:
        decoder.decode(42)
    assert str(exc_info.value) == (
        f"Argument for {type_name(MyClass)}."
        f"__mashumaro_from_dict__ method should be a dict instance"
    )


def test_extra_keys_error():
    @dataclass
    class MyClass(DataClassDictMixin):
        x: str

        class Config:
            forbid_extra_keys = True

    with pytest.raises(ExtraKeysError) as exc_info:
        MyClass.from_dict({"x": "x", "y": "y"})
    assert str(exc_info.value).endswith(".MyClass: y")



from unittest.mock import patch

from mashumaro.core.const import PY_312_MIN

if not PY_312_MIN:
    collect_ignore = [
        "test_generics_pep_695.py",
        "test_pep_695.py",
        "test_recursive_union.py",
    ]

add_unpack_method = patch(
    "mashumaro.core.meta.code.builder.CodeBuilder.add_unpack_method",
    lambda *args, **kwargs: ...,
)



from dataclasses import dataclass
from datetime import date

import pytest
from typing_extensions import Annotated, Literal

from mashumaro import DataClassDictMixin
from mashumaro.codecs.basic import decode
from mashumaro.exceptions import InvalidFieldValue
from mashumaro.types import Discriminator

DT_STR = "2023-05-30"
DT_DATE = date(2023, 5, 30)

X_STR_1 = {"x": DT_STR, "type": "str_1"}
X_STR_12 = {"x": DT_STR, "type": "str_12"}
X_DATE_12 = {"x": DT_STR, "type": "date_12"}
X_DATE_1 = {"x": DT_STR, "type": "date_1"}
X_STR_21 = {"x": DT_STR, "type": "str_21"}
X_DATE_22 = {"x": DT_STR, "type": "date_22"}


@dataclass
class BaseVariant(DataClassDictMixin):
    pass


@dataclass
class _BaseVariant:
    pass


@dataclass
class VariantStr1(BaseVariant):
    x: str
    type: Literal["str_1"] = "str_1"


@dataclass
class _VariantStr1(_BaseVariant):
    x: str
    type: Literal["str_1"] = "str_1"


@dataclass
class VariantDate1(BaseVariant):
    x: date
    type: Literal["date_1"] = "date_1"


@dataclass
class _VariantDate1(_BaseVariant):
    x: date
    type: Literal["date_1"] = "date_1"


@dataclass
class VariantStr12(VariantStr1):
    x: str
    type: Literal["str_12"] = "str_12"


@dataclass
class _VariantStr12(_VariantStr1):
    x: str
    type: Literal["str_12"] = "str_12"


@dataclass
class VariantDate12(VariantStr1):
    x: date
    type: Literal["date_12"] = "date_12"


@dataclass
class _VariantDate12(_VariantStr1):
    x: date
    type: Literal["date_12"] = "date_12"


@dataclass
class VariantStr21(VariantDate1):
    x: str
    type: Literal["str_21"] = "str_21"


@dataclass
class _VariantStr21(_VariantDate1):
    x: str
    type: Literal["str_21"] = "str_21"


@dataclass
class VariantDate22(VariantDate1):
    x: date
    type: Literal["date_22"] = "date_22"


@dataclass
class _VariantDate22(_VariantDate1):
    x: date
    type: Literal["date_22"] = "date_22"


@dataclass
class BySubtypes(DataClassDictMixin):
    x: Annotated[BaseVariant, Discriminator(include_subtypes=True)]


@dataclass
class _BySubtypes:
    x: Annotated[_BaseVariant, Discriminator(include_subtypes=True)]


@dataclass
class BySupertypes(DataClassDictMixin):
    x: Annotated[BaseVariant, Discriminator(include_supertypes=True)]
    x1: Annotated[VariantStr1, Discriminator(include_supertypes=True)]
    x2: Annotated[VariantDate1, Discriminator(include_supertypes=True)]


@dataclass
class _BySupertypes:
    x: Annotated[_BaseVariant, Discriminator(include_supertypes=True)]
    x1: Annotated[_VariantStr1, Discriminator(include_supertypes=True)]
    x2: Annotated[_VariantDate1, Discriminator(include_supertypes=True)]


@dataclass
class ByFieldWithSubtypes(DataClassDictMixin):
    x: Annotated[
        BaseVariant, Discriminator(field="type", include_subtypes=True)
    ]


@dataclass
class _ByFieldWithSubtypes:
    x: Annotated[
        _BaseVariant, Discriminator(field="type", include_subtypes=True)
    ]


@dataclass
class ByFieldWithSupertypes(DataClassDictMixin):
    x1: Annotated[
        VariantStr1, Discriminator(field="type", include_supertypes=True)
    ]
    x2: Annotated[
        VariantDate1, Discriminator(field="type", include_supertypes=True)
    ]


@dataclass
class _ByFieldWithSupertypes:
    x1: Annotated[
        _VariantStr1, Discriminator(field="type", include_supertypes=True)
    ]
    x2: Annotated[
        _VariantDate1, Discriminator(field="type", include_supertypes=True)
    ]


@dataclass
class ByFieldWithSupertypesAndSubtypes(DataClassDictMixin):
    x1: Annotated[
        VariantStr1,
        Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        ),
    ]
    x2: Annotated[
        VariantDate1,
        Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        ),
    ]


@dataclass
class _ByFieldWithSupertypesAndSubtypes:
    x1: Annotated[
        _VariantStr1,
        Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        ),
    ]
    x2: Annotated[
        _VariantDate1,
        Discriminator(
            field="type", include_supertypes=True, include_subtypes=True
        ),
    ]


@dataclass
class BySupertypesAndSubtypes(DataClassDictMixin):
    x1: Annotated[
        VariantStr1,
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]
    x2: Annotated[
        VariantDate1,
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]


@dataclass
class _BySupertypesAndSubtypes:
    x1: Annotated[
        _VariantStr1,
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]
    x2: Annotated[
        _VariantDate1,
        Discriminator(include_supertypes=True, include_subtypes=True),
    ]


@dataclass
class Foo1(DataClassDictMixin):
    x1: int


@dataclass
class _Foo1:
    x1: int


@dataclass
class Foo2(Foo1):
    x2: int


@dataclass
class _Foo2(_Foo1):
    x2: int


@dataclass
class Foo3(Foo2):
    x: int
    type = 3


@dataclass
class _Foo3(_Foo2):
    x: int
    type = 3


@dataclass
class Foo4(Foo2):
    x: int
    type = 4


@dataclass
class _Foo4(_Foo2):
    x: int
    type = 4


@dataclass
class Bar(DataClassDictMixin):
    baz1: Annotated[Foo1, Discriminator(field="type", include_subtypes=True)]
    baz2: Annotated[Foo1, Discriminator(include_subtypes=True)]


@dataclass
class _Bar:
    baz1: Annotated[_Foo1, Discriminator(field="type", include_subtypes=True)]
    baz2: Annotated[_Foo1, Discriminator(include_subtypes=True)]


@dataclass
class BaseVariantWitCustomTagger(DataClassDictMixin):
    pass


@dataclass
class _BaseVariantWitCustomTagger:
    pass


@dataclass
class VariantWitCustomTaggerSub1(BaseVariantWitCustomTagger):
    pass


@dataclass
class _VariantWitCustomTaggerSub1(_BaseVariantWitCustomTagger):
    pass


@dataclass
class VariantWitCustomTaggerSub2(BaseVariantWitCustomTagger):
    pass


@dataclass
class _VariantWitCustomTaggerSub2(_BaseVariantWitCustomTagger):
    pass


@dataclass
class VariantWitCustomTaggerOwner(DataClassDictMixin):
    x: Annotated[
        BaseVariantWitCustomTagger,
        Discriminator(
            field="type",
            include_subtypes=True,
            variant_tagger_fn=lambda cls: cls.__name__.lower(),
        ),
    ]


@dataclass
class _VariantWitCustomTaggerOwner:
    x: Annotated[
        _BaseVariantWitCustomTagger,
        Discriminator(
            field="type",
            include_subtypes=True,
            variant_tagger_fn=lambda cls: cls.__name__.lower(),
        ),
    ]


@pytest.mark.parametrize(
    ["variant_data", "variant"],
    [
        (X_STR_1, VariantStr1(DT_STR)),
        (X_STR_12, VariantStr12(DT_STR)),
        (X_DATE_12, VariantDate12(DT_DATE)),
        (X_DATE_1, VariantDate1(DT_DATE)),
        (X_STR_21, VariantStr21(DT_STR)),
        (X_DATE_22, VariantDate22(DT_DATE)),
    ],
)
def test_by_subtypes(variant_data, variant):
    assert BySubtypes.from_dict({"x": variant_data}) == BySubtypes(variant)


@pytest.mark.parametrize(
    ["variant_data", "variant"],
    [
        (X_STR_1, _VariantStr1(DT_STR)),
        (X_STR_12, _VariantStr12(DT_STR)),
        (X_DATE_12, _VariantDate12(DT_DATE)),
        (X_DATE_1, _VariantDate1(DT_DATE)),
        (X_STR_21, _VariantStr21(DT_STR)),
        (X_DATE_22, _VariantDate22(DT_DATE)),
    ],
)
def test_by_subtypes_with_decoder(variant_data, variant):
    assert decode({"x": variant_data}, _BySubtypes) == _BySubtypes(variant)


def test_by_subtypes_exceptions():
    with pytest.raises(InvalidFieldValue):
        BySubtypes.from_dict({"x": {"type": "unknown"}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {"type": "unknown"}}, _BySubtypes)


def test_by_supertypes():
    assert BySupertypes.from_dict(
        {"x": {}, "x1": X_STR_1, "x2": X_DATE_1}
    ) == BySupertypes(
        BaseVariant(), VariantStr1(DT_STR), VariantDate1(DT_DATE)
    )
    assert decode(
        {"x": {}, "x1": X_STR_1, "x2": X_DATE_1}, _BySupertypes
    ) == _BySupertypes(
        _BaseVariant(), _VariantStr1(DT_STR), _VariantDate1(DT_DATE)
    )

    assert BySupertypes.from_dict(
        {"x": X_STR_1, "x1": X_STR_1, "x2": X_DATE_1}
    ) == BySupertypes(
        BaseVariant(), VariantStr1(DT_STR), VariantDate1(DT_DATE)
    )
    assert decode(
        {"x": X_STR_1, "x1": X_STR_1, "x2": X_DATE_1}, _BySupertypes
    ) == _BySupertypes(
        _BaseVariant(), _VariantStr1(DT_STR), _VariantDate1(DT_DATE)
    )

    with pytest.raises(InvalidFieldValue) as exc_info:
        BySupertypes.from_dict({"x": {}, "x1": X_STR_12, "x2": X_DATE_1})
    assert exc_info.value.field_name == "x1"

    with pytest.raises(InvalidFieldValue) as exc_info:
        decode({"x": {}, "x1": X_STR_12, "x2": X_DATE_1}, _BySupertypes)
    assert exc_info.value.field_name == "x1"

    with pytest.raises(InvalidFieldValue) as exc_info:
        BySupertypes.from_dict({"x": {}, "x1": X_STR_1, "x2": X_DATE_22})
    assert exc_info.value.field_name == "x2"

    with pytest.raises(InvalidFieldValue) as exc_info:
        decode({"x": {}, "x1": X_STR_1, "x2": X_DATE_22}, _BySupertypes)
    assert exc_info.value.field_name == "x2"


@pytest.mark.parametrize(
    ["variant_data", "variant"],
    [
        (X_STR_1, VariantStr1(DT_STR)),
        (X_STR_12, VariantStr12(DT_STR)),
        (X_DATE_12, VariantDate12(DT_DATE)),
        (X_DATE_1, VariantDate1(DT_DATE)),
        (X_STR_21, VariantStr21(DT_STR)),
        (X_DATE_22, VariantDate22(DT_DATE)),
    ],
)
def test_by_field_with_subtypes(variant_data, variant):
    assert ByFieldWithSubtypes.from_dict(
        {"x": variant_data}
    ) == ByFieldWithSubtypes(variant)


@pytest.mark.parametrize(
    ["variant_data", "variant"],
    [
        (X_STR_1, _VariantStr1(DT_STR)),
        (X_STR_12, _VariantStr12(DT_STR)),
        (X_DATE_12, _VariantDate12(DT_DATE)),
        (X_DATE_1, _VariantDate1(DT_DATE)),
        (X_STR_21, _VariantStr21(DT_STR)),
        (X_DATE_22, _VariantDate22(DT_DATE)),
    ],
)
def test_by_field_with_subtypes_with_decoder(variant_data, variant):
    assert decode(
        {"x": variant_data}, _ByFieldWithSubtypes
    ) == _ByFieldWithSubtypes(variant)


def test_by_field_with_subtypes_exceptions():
    with pytest.raises(InvalidFieldValue):
        ByFieldWithSubtypes.from_dict({"x": {"type": "unknown"}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {"type": "unknown"}}, _ByFieldWithSubtypes)


def test_by_field_with_supertypes():
    assert ByFieldWithSupertypes.from_dict(
        {"x1": X_STR_1, "x2": X_DATE_1}
    ) == ByFieldWithSupertypes(VariantStr1(DT_STR), VariantDate1(DT_DATE))
    assert decode(
        {"x1": X_STR_1, "x2": X_DATE_1}, _ByFieldWithSupertypes
    ) == _ByFieldWithSupertypes(_VariantStr1(DT_STR), _VariantDate1(DT_DATE))

    with pytest.raises(InvalidFieldValue) as exc_info:
        ByFieldWithSupertypes.from_dict({"x1": X_STR_12, "x2": X_DATE_1})
    assert exc_info.value.field_name == "x1"

    with pytest.raises(InvalidFieldValue) as exc_info:
        decode({"x1": X_STR_12, "x2": X_DATE_1}, _ByFieldWithSupertypes)
    assert exc_info.value.field_name == "x1"

    with pytest.raises(InvalidFieldValue) as exc_info:
        ByFieldWithSupertypes.from_dict({"x1": X_STR_1, "x2": X_DATE_22})
    assert exc_info.value.field_name == "x2"

    with pytest.raises(InvalidFieldValue) as exc_info:
        decode({"x1": X_STR_1, "x2": X_DATE_22}, _ByFieldWithSupertypes)
    assert exc_info.value.field_name == "x2"


@pytest.mark.parametrize(
    ["dataclass_cls"],
    [[ByFieldWithSupertypesAndSubtypes], [BySupertypesAndSubtypes]],
)
def test_by_field_with_supertypes_and_subtypes(dataclass_cls):
    assert dataclass_cls.from_dict(
        {"x1": X_STR_1, "x2": X_DATE_1}
    ) == dataclass_cls(VariantStr1(DT_STR), VariantDate1(DT_DATE))

    assert dataclass_cls.from_dict(
        {"x1": X_STR_12, "x2": X_STR_21}
    ) == dataclass_cls(VariantStr12(DT_STR), VariantStr21(DT_STR))

    assert dataclass_cls.from_dict(
        {"x1": X_DATE_12, "x2": X_DATE_22}
    ) == dataclass_cls(VariantDate12(DT_DATE), VariantDate22(DT_DATE))

    with pytest.raises(InvalidFieldValue):
        dataclass_cls.from_dict({"x1": X_DATE_1, "x2": X_STR_1})

    with pytest.raises(InvalidFieldValue):
        dataclass_cls.from_dict({"x1": X_STR_21, "x2": X_STR_12})

    with pytest.raises(InvalidFieldValue):
        dataclass_cls.from_dict({"x1": X_DATE_22, "x2": X_DATE_12})


@pytest.mark.parametrize(
    ["dataclass_cls"],
    [[_ByFieldWithSupertypesAndSubtypes], [_BySupertypesAndSubtypes]],
)
def test_by_field_with_supertypes_and_subtypes_with_decoder(dataclass_cls):
    assert decode(
        {"x1": X_STR_1, "x2": X_DATE_1}, dataclass_cls
    ) == dataclass_cls(_VariantStr1(DT_STR), _VariantDate1(DT_DATE))

    assert decode(
        {"x1": X_STR_12, "x2": X_STR_21}, dataclass_cls
    ) == dataclass_cls(_VariantStr12(DT_STR), _VariantStr21(DT_STR))

    assert decode(
        {"x1": X_DATE_12, "x2": X_DATE_22}, dataclass_cls
    ) == dataclass_cls(_VariantDate12(DT_DATE), _VariantDate22(DT_DATE))

    with pytest.raises(InvalidFieldValue):
        decode({"x1": X_DATE_1, "x2": X_STR_1}, dataclass_cls)

    with pytest.raises(InvalidFieldValue):
        decode({"x1": X_STR_21, "x2": X_STR_12}, dataclass_cls)

    with pytest.raises(InvalidFieldValue):
        decode({"x1": X_DATE_22, "x2": X_DATE_12}, dataclass_cls)


def test_subclass_tree_with_class_without_field():
    assert Bar.from_dict(
        {
            "baz1": {"type": 4, "x1": 1, "x2": 2, "x": 42},
            "baz2": {"type": 4, "x1": 1, "x2": 2, "x": 42},
        }
    ) == Bar(baz1=Foo4(1, 2, 42), baz2=Foo2(1, 2))
    assert decode(
        {
            "baz1": {"type": 4, "x1": 1, "x2": 2, "x": 42},
            "baz2": {"type": 4, "x1": 1, "x2": 2, "x": 42},
        },
        _Bar,
    ) == _Bar(baz1=_Foo4(1, 2, 42), baz2=_Foo2(1, 2))


def test_by_field_with_custom_variant_tagger():
    assert VariantWitCustomTaggerOwner.from_dict(
        {"x": {"type": "variantwitcustomtaggersub1"}}
    ) == VariantWitCustomTaggerOwner(VariantWitCustomTaggerSub1())
    assert decode(
        {"x": {"type": "_variantwitcustomtaggersub1"}},
        _VariantWitCustomTaggerOwner,
    ) == _VariantWitCustomTaggerOwner(_VariantWitCustomTaggerSub1())

    assert VariantWitCustomTaggerOwner.from_dict(
        {"x": {"type": "variantwitcustomtaggersub2"}}
    ) == VariantWitCustomTaggerOwner(VariantWitCustomTaggerSub2())
    assert decode(
        {"x": {"type": "_variantwitcustomtaggersub2"}},
        _VariantWitCustomTaggerOwner,
    ) == _VariantWitCustomTaggerOwner(_VariantWitCustomTaggerSub2())

    with pytest.raises(InvalidFieldValue):
        VariantWitCustomTaggerOwner.from_dict({"x": {"type": "unknown"}})
    with pytest.raises(InvalidFieldValue):
        decode({"x": {"type": "unknown"}}, _VariantWitCustomTaggerOwner)



from collections import namedtuple
from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum, Flag, IntEnum, IntFlag
from os import PathLike
from typing import Any, Generic, List, NewType, Optional, Union

try:
    from enum import StrEnum
except ImportError:  # pragma: no cover

    class StrEnum(str, Enum):
        pass


from typing_extensions import NamedTuple, ReadOnly, TypedDict, TypeVar

from mashumaro import DataClassDictMixin
from mashumaro.config import TO_DICT_ADD_OMIT_NONE_FLAG, BaseConfig
from mashumaro.types import GenericSerializableType, SerializableType

T = TypeVar("T")
TAny = TypeVar("TAny", bound=Any)
TInt = TypeVar("TInt", bound=int)
TDefaultInt = TypeVar("TDefaultInt", default=int)
TIntStr = TypeVar("TIntStr", int, str)
T_Optional_int = TypeVar("T_Optional_int", bound=Optional[int])


class MyEnum(Enum):
    a = "letter a"
    b = "letter b"


class MyStrEnum(str, Enum):
    a = "letter a"
    b = "letter b"


class MyNativeStrEnum(StrEnum):
    a = "letter a"
    b = "letter b"


class MyIntEnum(IntEnum):
    a = 1
    b = 2


class MyFlag(Flag):
    a = 1
    b = 2


class MyIntFlag(IntFlag):
    a = 1
    b = 2


class MyList(list):
    pass


@dataclass
class MyDataClass(DataClassDictMixin):
    a: int
    b: int


@dataclass(frozen=True)
class MyFrozenDataClass:
    x: int


class MutableString(SerializableType):
    def __init__(self, value: str):
        self.characters = [c for c in value]

    def _serialize(self) -> str:
        return str(self)

    @classmethod
    def _deserialize(cls, value: str) -> "MutableString":
        return MutableString(value)

    def __str__(self):
        return "".join(self.characters)

    def __eq__(self, other):
        return self.characters == other.characters


class GenericSerializableList(Generic[T], GenericSerializableType):
    def __init__(self, value: List[T]):
        self.value = value

    def _serialize(self, types):
        if types[0] == int:
            return [v + 2 for v in self.value]
        elif types[0] == str:
            return [f"_{v}" for v in self.value]

    @classmethod
    def _deserialize(cls, value, types):
        if types[0] == int:
            return GenericSerializableList([int(v) - 2 for v in value])
        elif types[0] == str:
            return GenericSerializableList([v[1:] for v in value])

    def __eq__(self, other):
        return self.value == other.value


class GenericSerializableWrapper(Generic[T], GenericSerializableType):
    def __init__(self, value: T):
        self.value = value

    def _serialize(self, types):
        if types[0] == date:
            return self.value.isoformat()

    @classmethod
    def _deserialize(cls, value, types):
        if types[0] == date:
            return GenericSerializableWrapper(date.fromisoformat(value))

    def __eq__(self, other):
        return self.value == other.value


class CustomPath(PathLike):
    def __init__(self, *args: str):
        self._path = "/".join(args)

    def __fspath__(self):
        return self._path

    def __eq__(self, other):
        return isinstance(other, CustomPath) and self._path == other._path


@dataclass
class MyDataClassWithUnion(DataClassDictMixin):
    a: Union[int, str]
    b: Union[MyEnum, int]


@dataclass
class MyDataClassWithOptional(DataClassDictMixin):
    a: Optional[int] = None
    b: Optional[int] = None


@dataclass
class MyDataClassWithOptionalAndOmitNoneFlag(DataClassDictMixin):
    a: Optional[int] = None
    b: Optional[int] = None

    class Config(BaseConfig):
        code_generation_options = [TO_DICT_ADD_OMIT_NONE_FLAG]


class ThirdPartyType:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return isinstance(other, ThirdPartyType) and self.value == other.value


@dataclass
class DataClassWithoutMixin:
    i: int


@dataclass
class SerializableTypeDataClass(SerializableType):
    a: int
    b: int

    def _serialize(self):
        return {"a": self.a + 1, "b": self.b + 1}

    @classmethod
    def _deserialize(cls, value):
        a = value.get("a") - 1
        b = value.get("b") - 1
        return cls(a, b)


@dataclass
class GenericSerializableTypeDataClass(GenericSerializableType):
    a: int
    b: int

    def _serialize(self, types):
        return {"a": self.a + 1, "b": self.b + 1}

    @classmethod
    def _deserialize(cls, value, types):
        a = value.get("a") - 1
        b = value.get("b") - 1
        return cls(a, b)


@dataclass
class MyGenericDataClass(Generic[T], DataClassDictMixin):
    x: T


class MyGenericList(List[T]):
    pass


class SerializableTypeGenericList(Generic[T], SerializableType):
    def __init__(self, value: List[T]):
        self.value = value

    def _serialize(self):
        return self.value

    @classmethod
    def _deserialize(cls, value):
        return SerializableTypeGenericList(value)

    def __eq__(self, other):
        return self.value == other.value


TMyDataClass = TypeVar("TMyDataClass", bound=MyDataClass)


class TypedDictRequiredKeys(TypedDict):
    int: int
    float: float


class TypedDictOptionalKeys(TypedDict, total=False):
    int: int
    float: float


class TypedDictRequiredAndOptionalKeys(TypedDictRequiredKeys, total=False):
    str: str


class TypedDictRequiredKeysWithOptional(TypedDict):
    x: Optional[int]
    y: int


class TypedDictOptionalKeysWithOptional(TypedDict, total=False):
    x: Optional[int]
    y: float


class TypedDictWithReadOnly(TypedDict):
    x: ReadOnly[int]


class GenericTypedDict(TypedDict, Generic[T]):
    x: T
    y: int


class MyNamedTuple(NamedTuple):
    i: int
    f: float


class MyNamedTupleWithDefaults(NamedTuple):
    i: int = 1
    f: float = 2.0


class MyNamedTupleWithOptional(NamedTuple):
    i: Optional[int]
    f: int


MyUntypedNamedTuple = namedtuple("MyUntypedNamedTuple", ("i", "f"))


MyUntypedNamedTupleWithDefaults = namedtuple(
    "MyUntypedNamedTupleWithDefaults",
    ("i", "f"),
    defaults=(1, 2.0),
)


class GenericNamedTuple(NamedTuple, Generic[T]):
    x: T
    y: int


MyDatetimeNewType = NewType("MyDatetimeNewType", datetime)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from mashumaro import DataClassDictMixin


@dataclass
class Node(DataClassDictMixin):
    next: Optional[Node] = None


def test_self_reference():
    assert Node.from_dict({}) == Node()
    assert Node.from_dict({"next": {}}) == Node(Node())
    assert Node().to_dict() == {"next": None}
    assert Node(Node()).to_dict() == {"next": {"next": None}}



from dataclasses import asdict

import pyperf

from benchmark.common import AbstractBenchmark
from benchmark.libs.mashumaro.common import BasicDecoder, DefaultDialect, Issue


class Benchmark(AbstractBenchmark):
    LIBRARY = "asdict"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.decoder = BasicDecoder(Issue, default_dialect=DefaultDialect)

    def warmup(self, data) -> None:
        asdict(self.decoder.decode(data))

    def run_dumper(self, data) -> pyperf.Benchmark:
        obj = self.decoder.decode(data)
        return self._bench_dumper_func(asdict, obj)



from dataclasses import dataclass, field
from typing import Optional, Union

import pytest
from typing_extensions import Literal

from mashumaro import DataClassDictMixin
from mashumaro.config import TO_DICT_ADD_OMIT_NONE_FLAG, BaseConfig
from mashumaro.exceptions import ExtraKeysError, InvalidFieldValue
from mashumaro.types import Discriminator, SerializationStrategy

from .entities import (
    MyDataClassWithOptional,
    MyDataClassWithOptionalAndOmitNoneFlag,
    MyNamedTuple,
    MyNamedTupleWithDefaults,
    MyUntypedNamedTuple,
    MyUntypedNamedTupleWithDefaults,
    TypedDictRequiredKeys,
)


@dataclass
class LazyCompilationDataClass(DataClassDictMixin):
    x: int

    class Config(BaseConfig):
        lazy_compilation = True


def test_debug_true_option(mocker):
    mocked_print = mocker.patch("builtins.print")

    @dataclass
    class _(DataClassDictMixin):
        union: Union[int, str, MyNamedTuple]
        typed_dict: TypedDictRequiredKeys
        named_tuple: MyNamedTupleWithDefaults
        literal: Literal[1, 2, 3]

        class Config(BaseConfig):
            debug = True

    mocked_print.assert_called()


def test_config_without_base_config_base(mocker):
    mocked_print = mocker.patch("builtins.print")

    @dataclass
    class _(DataClassDictMixin):
        x: Union[int, str]

        class Config:
            debug = True

    mocked_print.assert_called()


def test_debug_false_option(mocker):
    mocked_print = mocker.patch("builtins.print")

    @dataclass
    class _(DataClassDictMixin):
        x: Union[int, str]

        class Config(BaseConfig):
            debug = False

    mocked_print.assert_not_called()


def test_omit_none_code_generation_flag():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[int] = None

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_OMIT_NONE_FLAG]

    assert DataClass().to_dict() == {"x": None}
    assert DataClass().to_dict(omit_none=True) == {}


def test_no_omit_none_code_generation_flag():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[int] = None

    assert DataClass().to_dict() == {"x": None}
    with pytest.raises(TypeError):
        DataClass().to_dict(omit_none=True)


def test_omit_none_flag_for_inner_class_without_it():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[MyDataClassWithOptional] = None

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_OMIT_NONE_FLAG]

    assert DataClass().to_dict() == {"x": None}
    assert DataClass().to_dict(omit_none=True) == {}

    empty_x = MyDataClassWithOptional()
    assert DataClass(empty_x).to_dict() == {"x": {"a": None, "b": None}}
    assert DataClass(empty_x).to_dict(omit_none=True) == {
        "x": {"a": None, "b": None}
    }


def test_omit_none_flag_for_inner_class_with_it():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[MyDataClassWithOptionalAndOmitNoneFlag] = None

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_OMIT_NONE_FLAG]

    assert DataClass().to_dict() == {"x": None}
    assert DataClass().to_dict(omit_none=True) == {}

    empty_x = MyDataClassWithOptionalAndOmitNoneFlag()
    assert DataClass(empty_x).to_dict() == {"x": {"a": None, "b": None}}
    assert DataClass(empty_x).to_dict(omit_none=True) == {"x": {}}


def test_passing_omit_none_into_union():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Optional[int] = None
        b: Optional[Union[int, MyDataClassWithOptionalAndOmitNoneFlag]] = None

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_OMIT_NONE_FLAG]

    instance = DataClass(b=MyDataClassWithOptionalAndOmitNoneFlag(a=1))
    assert instance.to_dict(omit_none=True) == {"b": {"a": 1}}


def test_serialization_strategy():
    class TestSerializationStrategy(SerializationStrategy):
        def serialize(self, value):
            return [value]

        def deserialize(self, value):
            return value[0]

    @dataclass
    class DataClass(DataClassDictMixin):
        a: int
        b: str
        c: int

        class Config(BaseConfig):
            serialization_strategy = {
                int: TestSerializationStrategy(),
                str: {
                    "serialize": lambda v: [v],
                    "deserialize": lambda v: v[0],
                },
            }

    obj = DataClass(a=123, b="abc", c=123)
    assert DataClass.from_dict({"a": [123], "b": ["abc"], "c": [123]}) == obj
    assert obj.to_dict() == {"a": [123], "b": ["abc"], "c": [123]}


def test_named_tuple_as_dict():
    @dataclass
    class DataClass(DataClassDictMixin):
        mnp: MyNamedTuple
        mnpwd: MyNamedTupleWithDefaults
        munp: MyUntypedNamedTuple

        class Config(BaseConfig):
            namedtuple_as_dict = True

    obj = DataClass(
        mnp=MyNamedTuple(i=1, f=2.0),
        mnpwd=MyNamedTupleWithDefaults(i=1, f=2.0),
        munp=MyUntypedNamedTuple(i=1, f=2.0),
    )
    obj_dict = {
        "mnp": {"i": 1, "f": 2.0},
        "mnpwd": {"i": 1, "f": 2.0},
        "munp": {"i": 1, "f": 2.0},
    }
    assert obj.to_dict() == obj_dict
    assert DataClass.from_dict(obj_dict) == obj


def test_untyped_named_tuple_with_defaults_as_dict():
    @dataclass
    class DataClass(DataClassDictMixin):
        munpwd: MyUntypedNamedTupleWithDefaults

        class Config(BaseConfig):
            namedtuple_as_dict = True

    obj = DataClass(munpwd=MyUntypedNamedTupleWithDefaults(i=1, f=2.0))
    assert obj.to_dict() == {"munpwd": {"i": 1, "f": 2.0}}
    assert DataClass.from_dict({"munpwd": {"i": 1, "f": 2.0}}) == obj


def test_named_tuple_as_dict_and_as_list_engine():
    @dataclass
    class DataClass(DataClassDictMixin):
        as_dict: MyNamedTuple
        as_list: MyNamedTuple = field(
            metadata={"serialize": "as_list", "deserialize": "as_list"}
        )

        class Config(BaseConfig):
            namedtuple_as_dict = True

    obj = DataClass(
        as_dict=MyNamedTuple(i=1, f=2.0),
        as_list=MyNamedTuple(i=1, f=2.0),
    )
    obj_dict = {
        "as_dict": {"i": 1, "f": 2.0},
        "as_list": [1, 2.0],
    }
    assert obj.to_dict() == obj_dict
    assert DataClass.from_dict(obj_dict) == obj


def test_omit_none_code_generation_flag_with_omit_none_by_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[int] = None

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_OMIT_NONE_FLAG]
            omit_none = True

    assert DataClass().to_dict() == {}
    assert DataClass().to_dict(omit_none=True) == {}
    assert DataClass().to_dict(omit_none=False) == {"x": None}


def test_lazy_compilation():
    obj = LazyCompilationDataClass(42)
    assert LazyCompilationDataClass.from_dict({"x": "42"}) == obj
    assert obj.to_dict() == {"x": 42}


def test_sort_keys_with_mixin():
    @dataclass
    class SortedDataClass(DataClassDictMixin):
        foo: int
        bar: int

        class Config(BaseConfig):
            sort_keys = True

    @dataclass
    class UnSortedDataClass(DataClassDictMixin):
        foo: int
        bar: int

        class Config(BaseConfig):
            sort_keys = False

    t = SortedDataClass(1, 2)
    assert str(t.to_dict()) == "{'bar': 2, 'foo': 1}"

    t = UnSortedDataClass(1, 2)
    assert str(t.to_dict()) == "{'foo': 1, 'bar': 2}"


@dataclass
class SortedDataClass:
    foo: int
    bar: int

    class Config(BaseConfig):
        sort_keys = True


@dataclass
class UnSortedDataClass:
    foo: int
    bar: int

    class Config(BaseConfig):
        sort_keys = False


def test_sort_keys_plain_dataclass():
    @dataclass
    class RootSortedModel(DataClassDictMixin):
        unsorted_sub: UnSortedDataClass
        sorted_sub: SortedDataClass

        class Config(BaseConfig):
            sort_keys = True

    @dataclass
    class RootUnSortedModel(DataClassDictMixin):
        unsorted_sub: UnSortedDataClass
        sorted_sub: SortedDataClass

        class Config(BaseConfig):
            sort_keys = False

    t = RootSortedModel(
        unsorted_sub=UnSortedDataClass(1, 2),
        sorted_sub=SortedDataClass(1, 2),
    )
    assert str(t.to_dict()) == (
        "{'sorted_sub': {'bar': 2, 'foo': 1}, "
        "'unsorted_sub': {'foo': 1, 'bar': 2}}"
    )

    t = RootUnSortedModel(
        unsorted_sub=UnSortedDataClass(1, 2),
        sorted_sub=SortedDataClass(1, 2),
    )
    assert str(t.to_dict()) == (
        "{'unsorted_sub': {'foo': 1, 'bar': 2}, "
        "'sorted_sub': {'bar': 2, 'foo': 1}}"
    )


def test_forbid_extra_keys():
    @dataclass
    class ForbidKeysModel(DataClassDictMixin):
        foo: int

        class Config(BaseConfig):
            forbid_extra_keys = True

    # Test unpacking works
    assert ForbidKeysModel.from_dict({"foo": 1}) == ForbidKeysModel(1)

    # Test extra keys are forbidden
    with pytest.raises(ExtraKeysError) as exc_info:
        ForbidKeysModel.from_dict({"foo": 1, "bar": 2, "baz": 3})

    assert exc_info.value.extra_keys == {"bar", "baz"}
    assert exc_info.value.target_type == ForbidKeysModel

    # Now with alias, but not allow_deserialization_not_by_alias
    @dataclass
    class ForbidKeysModel(DataClassDictMixin):
        foo: int = field(metadata={"alias": "f"})
        bar: int

        class Config(BaseConfig):
            forbid_extra_keys = True
            aliases = {"bar": "b"}

    # Test unpacking works
    assert ForbidKeysModel.from_dict({"f": 1, "b": 2}) == ForbidKeysModel(1, 2)

    # Test extra keys are forbidden
    with pytest.raises(ExtraKeysError) as exc_info:
        ForbidKeysModel.from_dict({"foo": 1, "bar": 2})

    assert exc_info.value.extra_keys == {"foo", "bar"}
    assert exc_info.value.target_type == ForbidKeysModel

    # Now with alias, but allow_deserialization_not_by_alias
    @dataclass
    class ForbidKeysModel(DataClassDictMixin):
        foo: int = field(metadata={"alias": "f"})
        bar: int

        class Config(BaseConfig):
            forbid_extra_keys = True
            aliases = {"bar": "b"}
            allow_deserialization_not_by_alias = True

    # Test unpacking works
    assert ForbidKeysModel.from_dict({"f": 1, "bar": 2}) == ForbidKeysModel(
        1, 2
    )

    # Test extra keys are forbidden
    with pytest.raises(ExtraKeysError) as exc_info:
        ForbidKeysModel.from_dict({"foo": 1, "b": 2, "baz": 3})

    assert exc_info.value.extra_keys == {"baz"}
    assert exc_info.value.target_type == ForbidKeysModel


@dataclass
class _VariantByBase(DataClassDictMixin):
    class Config(BaseConfig):
        discriminator = Discriminator(
            field="__type",
            include_subtypes=True,
            variant_tagger_fn=lambda clz: clz.__name__,
        )
        forbid_extra_keys = True


@dataclass
class _VariantByField1(_VariantByBase):
    x: Optional[str] = None


@dataclass
class _VariantByField2(_VariantByBase):
    x: Optional[str] = None


@dataclass
class _VariantByField3(_VariantByBase):
    x: Optional[str] = None

    class Config(BaseConfig):
        # add intermediate config to check if we're looking for the
        # discriminator in the base class _VariantByBase to add the
        # discriminator field to the list of allowed keys for _VariantByField4
        # deserializable via _VariantByBase.from_dict
        omit_none = True


@dataclass
class _VariantByField4(_VariantByField3):
    class Config(BaseConfig):
        forbid_extra_keys = True


@dataclass
class ForbidKeysModelWithDiscriminator(DataClassDictMixin):
    inner: _VariantByBase

    class Config(BaseConfig):
        forbid_extra_keys = True


def test_forbid_extra_keys_with_discriminator():
    # Test unpacking works
    assert ForbidKeysModelWithDiscriminator.from_dict(
        {"inner": {"x": "foo", "__type": "_VariantByField2"}}
    ) == ForbidKeysModelWithDiscriminator(_VariantByField2(x="foo"))

    # Test extra keys are forbidden
    with pytest.raises(InvalidFieldValue) as exc_info:
        ForbidKeysModelWithDiscriminator.from_dict(
            {"inner": {"x": "foo", "__type": "_VariantByField2", "bar": "baz"}}
        )

    root_exc = exc_info.value.__context__
    assert isinstance(root_exc, ExtraKeysError)
    assert root_exc.extra_keys == {"bar"}
    assert root_exc.target_type == _VariantByField2


def test_forbid_extra_keys_with_discriminator_for_subclass():
    assert _VariantByBase.from_dict(
        {"x": "foo", "__type": "_VariantByField4"}
    ) == _VariantByField4("foo")
    with pytest.raises(ExtraKeysError) as exc_info:
        _VariantByBase.from_dict(
            {"x": "foo", "__type": "_VariantByField4", "y": "bar"}
        )
    assert exc_info.value.extra_keys == {"y"}
